from unittest import IsolatedAsyncioTestCase

import pytest

from ...utils.testing import create_test_profile
from ...wallet.base import BaseWallet
from ...wallet.key_type import BLS12381G2
from ..ld_proofs import (
    AssertionProofPurpose,
    BbsBlsSignature2020,
    BbsBlsSignatureProof2020,
    WalletKeyPair,
    derive,
    sign,
    verify,
)
from .data import (
    BBS_NESTED_VC_FULL_REVEAL_DOCUMENT_MATTR,
    BBS_NESTED_VC_MATTR,
    BBS_NESTED_VC_REVEAL_DOCUMENT_MATTR,
    BBS_PARTIAL_PROOF_NESTED_VC_MATTR,
    BBS_PARTIAL_PROOF_VC_MATTR,
    BBS_PROOF_NESTED_VC_MATTR,
    BBS_PROOF_VC_MATTR,
    BBS_SIGNED_NESTED_VC_MATTR,
    BBS_SIGNED_VC_MATTR,
    BBS_VC_MATTR,
    BBS_VC_REVEAL_DOCUMENT_MATTR,
)
from .document_loader import custom_document_loader


@pytest.mark.ursa_bbs_signatures
class TestBbsMattrInterop(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            key_info = await wallet.create_key(
                key_type=BLS12381G2,
            )

        self.signature_issuer_suite = BbsBlsSignature2020(
            verification_method="did:example:489398593#test",
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=BLS12381G2,
                public_key_base58=key_info.verkey,
            ),
        )

        self.signature_suite = BbsBlsSignature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=BLS12381G2),
        )
        self.proof_suite = BbsBlsSignatureProof2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=BLS12381G2)
        )

    async def test_sign_bbs_vc_mattr(self):
        # TODO: ideally we verify the resulting document with the
        # MATTR bbs-jsonld lib. This is now manually done.
        signed = await sign(
            document=BBS_VC_MATTR,
            suite=self.signature_issuer_suite,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert signed

    async def test_sign_bbs_nested_vc_mattr(self):
        # TODO: ideally we verify the resulting document with the
        # MATTR bbs-jsonld lib. This is now manually done.
        signed = await sign(
            document=BBS_NESTED_VC_MATTR,
            suite=self.signature_issuer_suite,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert signed

    async def test_verify_bbs_signed_vc_mattr(self):
        result = await verify(
            document=BBS_SIGNED_VC_MATTR,
            suites=[self.signature_suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified

    async def test_verify_bbs_signed_nested_vc_mattr(self):
        result = await verify(
            document=BBS_SIGNED_NESTED_VC_MATTR,
            suites=[self.signature_suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified

    async def test_derive_bbs_proof_vc_mattr(self):
        # TODO: ideally we verify the resulting document with the
        # MATTR bbs-jsonld lib. This is now manually done.
        derived = await derive(
            document=BBS_SIGNED_VC_MATTR,
            reveal_document=BBS_VC_REVEAL_DOCUMENT_MATTR,
            suite=self.proof_suite,
            document_loader=custom_document_loader,
        )

        assert derived

    async def test_derive_full_bbs_proof_nested_vc_mattr(self):
        # TODO: ideally we verify the resulting document with the
        # MATTR bbs-jsonld lib. This is now manually done.
        derived = await derive(
            document=BBS_SIGNED_NESTED_VC_MATTR,
            reveal_document=BBS_NESTED_VC_FULL_REVEAL_DOCUMENT_MATTR,
            suite=self.proof_suite,
            document_loader=custom_document_loader,
        )

        assert derived

    async def test_derive_partial_bbs_proof_nested_vc_mattr(self):
        # TODO: ideally we verify the resulting document with the
        # MATTR bbs-jsonld lib. This is now manually done.
        derived = await derive(
            document=BBS_SIGNED_NESTED_VC_MATTR,
            reveal_document=BBS_NESTED_VC_REVEAL_DOCUMENT_MATTR,
            suite=self.proof_suite,
            document_loader=custom_document_loader,
        )

        assert derived

    async def test_verify_bbs_proof_vc_mattr(self):
        # BBS full proof, no subject nesting
        result = await verify(
            document=BBS_PROOF_VC_MATTR,
            suites=[self.proof_suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified

    async def test_verify_bbs_partial_proof_vc_mattr(self):
        # BBS partial proof, no subject nesting
        result = await verify(
            document=BBS_PARTIAL_PROOF_VC_MATTR,
            suites=[self.proof_suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified

    async def test_verify_bbs_proof_nested_vc_mattr(self):
        # BBS full proof nested vc
        result = await verify(
            document=BBS_PROOF_NESTED_VC_MATTR,
            suites=[self.proof_suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified

    async def test_verify_bbs_partial_proof_nested_vc_mattr(self):
        # BBS partial proof nested vc
        result = await verify(
            document=BBS_PARTIAL_PROOF_NESTED_VC_MATTR,
            suites=[self.proof_suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified



from unittest import IsolatedAsyncioTestCase

from aiohttp import web

from ...tests import mock
from ...utils.testing import create_test_profile
from ..decorators.auth import admin_authentication, tenant_authentication
from ..request_context import AdminRequestContext


class TestAdminAuthentication(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "admin_api_key",
                "admin.admin_insecure_mode": False,
            }
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k], headers={}, method="POST"
        )
        self.decorated_handler = mock.CoroutineMock()

    async def test_options_request(self):
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k], headers={}, method="OPTIONS"
        )
        decor_func = admin_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_insecure_mode(self):
        self.profile.settings["admin.admin_insecure_mode"] = True
        decor_func = admin_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_invalid_api_key(self):
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "wrong-key"},
            method="POST",
        )
        decor_func = admin_authentication(self.decorated_handler)
        with self.assertRaises(web.HTTPUnauthorized):
            await decor_func(self.request)

    async def test_valid_api_key(self):
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "admin_api_key"},
            method="POST",
        )
        decor_func = admin_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)


class TestTenantAuthentication(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "admin_api_key",
                "admin.admin_insecure_mode": False,
                "multitenant.enabled": True,
            }
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k], headers={}, method="POST"
        )
        self.decorated_handler = mock.CoroutineMock()

    async def test_options_request(self):
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k], headers={}, method="OPTIONS"
        )
        decor_func = tenant_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_insecure_mode_witout_token(self):
        self.profile.settings["admin.admin_insecure_mode"] = True
        decor_func = tenant_authentication(self.decorated_handler)
        with self.assertRaises(web.HTTPUnauthorized):
            await decor_func(self.request)

    async def test_single_tenant_invalid_api_key(self):
        self.profile.settings["multitenant.enabled"] = False
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "wrong-key"},
            method="POST",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        with self.assertRaises(web.HTTPUnauthorized):
            await decor_func(self.request)

    async def test_single_tenant_valid_api_key(self):
        self.profile.settings["multitenant.enabled"] = False
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "admin_api_key"},
            method="POST",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_multi_tenant_missing_auth_header(self):
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "wrong-key"},
            method="POST",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        with self.assertRaises(web.HTTPUnauthorized):
            await decor_func(self.request)

    async def test_multi_tenant_valid_auth_header(self):
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "admin_api_key", "Authorization": "Bearer my-jwt"},
            method="POST",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_base_wallet_additional_route_allowed_string(self):
        self.profile.settings["multitenant.base_wallet_routes"] = (
            "/not-this-route /extra-route"
        )
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "admin_api_key"},
            method="POST",
            path="/extra-route",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_base_wallet_additional_route_allowed_list(self):
        self.profile.settings["multitenant.base_wallet_routes"] = [
            "/extra-route",
            "/not-this-route",
        ]
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "admin_api_key"},
            method="POST",
            path="/extra-route",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        await decor_func(self.request)
        self.decorated_handler.assert_called_once_with(self.request)

    async def test_base_wallet_additional_route_denied(self):
        self.profile.settings["multitenant.base_wallet_routes"] = "/extra-route"
        self.request = mock.MagicMock(
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "admin_api_key"},
            method="POST",
            path="/extra-route-wrong",
        )
        decor_func = tenant_authentication(self.decorated_handler)
        with self.assertRaises(web.HTTPUnauthorized):
            await decor_func(self.request)



"""Feature discovery v2 admin routes."""

from aiohttp import web
from aiohttp_apispec import docs, querystring_schema, response_schema
from marshmallow import fields

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUID4_EXAMPLE
from ....storage.error import StorageError, StorageNotFoundError
from .manager import V20DiscoveryMgr
from .message_types import SPEC_URI
from .models.discovery_record import V20DiscoveryExchangeRecord, V20DiscoveryRecordSchema


class V20DiscoveryExchangeResultSchema(OpenAPISchema):
    """Result schema for Discover Features v2.0 exchange record."""

    results = fields.Nested(
        V20DiscoveryRecordSchema,
        metadata={"description": "Discover Features v2.0 exchange record"},
    )


class V20DiscoveryExchangeListResultSchema(OpenAPISchema):
    """Result schema for Discover Features v2.0 exchange records."""

    results = fields.List(
        fields.Nested(
            V20DiscoveryRecordSchema,
            metadata={"description": "Discover Features v2.0 exchange record"},
        )
    )


class QueryFeaturesQueryStringSchema(OpenAPISchema):
    """Query string parameters for feature query."""

    query_protocol = fields.Str(
        required=False,
        metadata={"description": "Protocol feature-type query", "example": "*"},
    )
    query_goal_code = fields.Str(
        required=False,
        metadata={"description": "Goal-code feature-type query", "example": "*"},
    )
    connection_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "Connection identifier, if none specified, then the query will provide"
                " features for this agent."
            ),
            "example": UUID4_EXAMPLE,
        },
    )


class QueryDiscoveryExchRecordsSchema(OpenAPISchema):
    """Query string parameter for Discover Features v2.0 exchange record."""

    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


@docs(
    tags=["discover-features v2.0"],
    summary="Query supported features",
)
@querystring_schema(QueryFeaturesQueryStringSchema())
@response_schema(V20DiscoveryExchangeResultSchema(), 200, description="")
@tenant_authentication
async def query_features(request: web.BaseRequest):
    """Request handler for creating and sending feature queries.

    Args:
        request: aiohttp request object

    Returns:
        V20DiscoveryExchangeRecord

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    mgr = V20DiscoveryMgr(profile)
    query_protocol = request.query.get("query_protocol", "*")
    query_goal_code = request.query.get("query_goal_code", "*")
    connection_id = request.query.get("connection_id")
    result = await mgr.create_and_send_query(
        connection_id=connection_id,
        query_protocol=query_protocol,
        query_goal_code=query_goal_code,
    )
    return web.json_response(result.serialize())


@docs(
    tags=["discover-features v2.0"],
    summary="Discover Features v2.0 records",
)
@querystring_schema(QueryDiscoveryExchRecordsSchema())
@response_schema(V20DiscoveryExchangeListResultSchema(), 200, description="")
@tenant_authentication
async def query_records(request: web.BaseRequest):
    """Request handler for looking up V20DiscoveryExchangeRecord.

    Args:
        request: aiohttp request object

    Returns:
        List of V20DiscoveryExchangeRecord

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.query.get("connection_id")
    if not connection_id:
        try:
            async with context.profile.session() as session:
                records = await V20DiscoveryExchangeRecord.query(session=session)
            results = [record.serialize() for record in records]
        except (StorageError, BaseModelError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err
    else:
        try:
            async with context.profile.session() as session:
                record = await V20DiscoveryExchangeRecord.retrieve_by_connection_id(
                    session=session, connection_id=connection_id
                )
            # There should only be one record for a connection
            results = [record.serialize()]
        except (StorageError, BaseModelError, StorageNotFoundError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response({"results": results})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/discover-features-2.0/queries", query_features, allow_head=False),
            web.get("/discover-features-2.0/records", query_records, allow_head=False),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "discover-features v2.0",
            "description": "Feature discovery v2",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""Credential exchange admin routes."""

import logging
from json.decoder import JSONDecodeError
from typing import Mapping, Optional

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import ValidationError, fields, validate, validates_schema

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....anoncreds.holder import AnonCredsHolderError
from ....anoncreds.issuer import AnonCredsIssuerError
from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile
from ....indy.holder import IndyHolderError
from ....indy.issuer import IndyIssuerError
from ....ledger.error import LedgerError
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.models.paginated_query import (
    PaginatedQuerySchema,
    get_paginated_query_params,
)
from ....messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_DID_EXAMPLE,
    ANONCREDS_SCHEMA_ID_EXAMPLE,
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
)
from ....storage.error import StorageError, StorageNotFoundError
from ....utils.tracing import AdminAPIMessageTracingSchema, get_timer, trace_event
from ....vc.ld_proofs.error import LinkedDataProofException
from ....wallet.util import default_did_from_verkey
from ...out_of_band.v1_0.models.oob_record import OobRecord
from . import problem_report_for_record, report_problem
from .formats.handler import V20CredFormatError
from .formats.ld_proof.models.cred_detail import LDProofVCDetailSchema
from .manager import V20CredManager, V20CredManagerError
from .message_types import ATTACHMENT_FORMAT, CRED_20_PROPOSAL, SPEC_URI
from .messages.cred_format import V20CredFormat
from .messages.cred_problem_report import ProblemReportReason
from .messages.cred_proposal import V20CredProposal
from .messages.inner.cred_preview import V20CredPreview, V20CredPreviewSchema
from .models.cred_ex_record import V20CredExRecord, V20CredExRecordSchema
from .models.detail.anoncreds import V20CredExRecordAnonCredsSchema
from .models.detail.indy import V20CredExRecordIndySchema
from .models.detail.ld_proof import V20CredExRecordLDProofSchema

LOGGER = logging.getLogger(__name__)


class V20IssueCredentialModuleResponseSchema(OpenAPISchema):
    """Response schema for v2.0 Issue Credential Module."""


class V20CredExRecordListQueryStringSchema(PaginatedQuerySchema):
    """Parameters and validators for credential exchange record list query."""

    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V20CredExRecord, m)
                for m in vars(V20CredExRecord)
                if m.startswith("ROLE_")
            ]
        ),
        metadata={"description": "Role assigned in credential exchange"},
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V20CredExRecord, m)
                for m in vars(V20CredExRecord)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Credential exchange state"},
    )


class V20CredExRecordDetailSchema(OpenAPISchema):
    """Credential exchange record and any per-format details."""

    cred_ex_record = fields.Nested(
        V20CredExRecordSchema,
        required=False,
        metadata={"description": "Credential exchange record"},
    )
    anoncreds = fields.Nested(V20CredExRecordAnonCredsSchema, required=False)
    indy = fields.Nested(V20CredExRecordIndySchema, required=False)
    ld_proof = fields.Nested(V20CredExRecordLDProofSchema, required=False)
    vc_di = fields.Nested(V20CredExRecordSchema, required=False)


class V20CredExRecordListResultSchema(OpenAPISchema):
    """Result schema for credential exchange record list query."""

    results = fields.List(
        fields.Nested(V20CredExRecordDetailSchema),
        metadata={
            "description": (
                "Credential exchange records and corresponding detail records"
            )
        },
    )


class V20CredStoreRequestSchema(OpenAPISchema):
    """Request schema for sending a credential store admin message."""

    credential_id = fields.Str(required=False)


class V20CredFilterAnonCredsSchema(OpenAPISchema):
    """AnonCreds credential filtration criteria."""

    schema_issuer_id = fields.Str(
        required=False,
        metadata={
            "description": "Schema issuer ID",
            "example": ANONCREDS_DID_EXAMPLE,
        },
    )
    schema_name = fields.Str(
        required=False,
        metadata={"description": "Schema name", "example": "preferences"},
    )
    schema_version = fields.Str(
        required=False,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    schema_id = fields.Str(
        required=False,
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
    )
    issuer_id = fields.Str(
        required=False,
        metadata={
            "description": "Credential issuer ID",
            "example": ANONCREDS_DID_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        required=False,
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )


class V20CredFilterIndySchema(OpenAPISchema):
    """Indy credential filtration criteria."""

    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    schema_id = fields.Str(
        required=False,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Schema issuer DID", "example": INDY_DID_EXAMPLE},
    )
    schema_name = fields.Str(
        required=False,
        metadata={"description": "Schema name", "example": "preferences"},
    )
    schema_version = fields.Str(
        required=False,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Credential issuer DID", "example": INDY_DID_EXAMPLE},
    )


class V20CredFilterVCDISchema(OpenAPISchema):
    """VCDI credential filtration criteria."""

    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    schema_id = fields.Str(
        required=False,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Schema issuer DID", "example": INDY_DID_EXAMPLE},
    )
    schema_name = fields.Str(
        required=False,
        metadata={"description": "Schema name", "example": "preferences"},
    )
    schema_version = fields.Str(
        required=False,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Credential issuer DID", "example": INDY_DID_EXAMPLE},
    )


class V20CredFilterSchema(OpenAPISchema):
    """Credential filtration criteria."""

    anoncreds = fields.Nested(
        V20CredFilterAnonCredsSchema,
        required=False,
        metadata={"description": "Credential filter for anoncreds"},
    )
    indy = fields.Nested(
        V20CredFilterIndySchema,
        required=False,
        metadata={"description": "Credential filter for indy"},
    )
    ld_proof = fields.Nested(
        LDProofVCDetailSchema,
        required=False,
        metadata={"description": "Credential filter for linked data proof"},
    )
    vc_di = fields.Nested(
        V20CredFilterVCDISchema,
        required=False,
        metadata={"description": "Credential filter for vc_di"},
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Data must have indy, ld_proof, vc_di, or all.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data has neither indy nor ld_proof

        """
        if not any(f.api in data for f in V20CredFormat.Format):
            raise ValidationError(
                "V20CredFilterSchema requires indy, ld_proof, vc_di or all"
            )


class V20IssueCredSchemaCore(AdminAPIMessageTracingSchema):
    """Filter, auto-remove, comment, trace."""

    filter_ = fields.Nested(
        V20CredFilterSchema,
        required=True,
        data_key="filter",
        metadata={"description": "Credential specification criteria by format"},
    )
    auto_remove = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )

    credential_preview = fields.Nested(V20CredPreviewSchema, required=False)

    replacement_id = fields.Str(
        required=False,
        allow_none=True,
        metadata={
            "description": "Optional identifier used to manage credential replacement",
            "example": UUID4_EXAMPLE,
        },
    )

    @validates_schema
    def validate(self, data, **kwargs):
        """Make sure preview is present when indy/vc_di format is present."""

        if (
            data.get("filter", {}).get("indy") or data.get("filter", {}).get("vc_di")
        ) and not data.get("credential_preview"):
            raise ValidationError(
                "Credential preview is required if indy or vc_di filter is present"
            )


class V20CredFilterLDProofSchema(OpenAPISchema):
    """Credential filtration criteria."""

    ld_proof = fields.Nested(
        LDProofVCDetailSchema,
        required=True,
        metadata={"description": "Credential filter for linked data proof"},
    )


class V20CredRequestFreeSchema(AdminAPIMessageTracingSchema):
    """Filter, auto-remove, comment, trace."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    # Request can only start with LD Proof
    filter_ = fields.Nested(
        V20CredFilterLDProofSchema,
        required=True,
        data_key="filter",
        metadata={"description": "Credential specification criteria by format"},
    )
    auto_remove = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    trace = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to trace event (default false)",
            "example": False,
        },
    )
    holder_did = fields.Str(
        required=False,
        allow_none=True,
        metadata={
            "description": "Holder DID to substitute for the credentialSubject.id",
            "example": "did:key:ahsdkjahsdkjhaskjdhakjshdkajhsdkjahs",
        },
    )


class V20CredExFreeSchema(V20IssueCredSchemaCore):
    """Request schema for sending credential admin message."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )

    verification_method = fields.Str(
        required=False,
        dump_default=None,
        allow_none=True,
        metadata={"description": "For ld-proofs. Verification method for signing."},
    )


class V20CredBoundOfferRequestSchema(OpenAPISchema):
    """Request schema for sending bound credential offer admin message."""

    filter_ = fields.Nested(
        V20CredFilterSchema,
        required=False,
        data_key="filter",
        metadata={"description": "Credential specification criteria by format"},
    )
    counter_preview = fields.Nested(
        V20CredPreviewSchema,
        required=False,
        metadata={"description": "Optional content for counter-proposal"},
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields: need both filter and counter_preview or neither."""
        if (
            "filter_" in data
            and ("indy" in data["filter_"] or "ld_proof" in data["filter_"])
        ) ^ ("counter_preview" in data):
            raise ValidationError(
                f"V20CredBoundOfferRequestSchema\n{data}\nrequires "
                "both indy/ld_proof filter and counter_preview or neither"
            )


class V20CredOfferRequestSchema(V20IssueCredSchemaCore):
    """Request schema for sending credential offer admin message."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    auto_issue = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to respond automatically to credential requests, creating and"
                " issuing requested credentials"
            )
        },
    )


class V20CredOfferConnFreeRequestSchema(V20IssueCredSchemaCore):
    """Request schema for creating credential offer free from connection."""

    auto_issue = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to respond automatically to credential requests, creating and"
                " issuing requested credentials"
            )
        },
    )


class V20CredRequestRequestSchema(OpenAPISchema):
    """Request schema for sending credential request message."""

    holder_did = fields.Str(
        required=False,
        allow_none=True,
        metadata={
            "description": "Holder DID to substitute for the credentialSubject.id",
            "example": "did:key:ahsdkjahsdkjhaskjdhakjshdkajhsdkjahs",
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )


class V20CredIssueRequestSchema(OpenAPISchema):
    """Request schema for sending credential issue admin message."""

    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )


class V20CredIssueProblemReportRequestSchema(OpenAPISchema):
    """Request schema for sending problem report."""

    description = fields.Str(required=True)


class V20CredIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential id."""

    credential_id = fields.Str(
        required=True,
        metadata={"description": "Credential identifier", "example": UUID4_EXAMPLE},
    )


class V20CredExIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential exchange id."""

    cred_ex_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Credential exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )


def _formats_filters(filt_spec: Mapping) -> Mapping:
    """Break out formats and filters for v2.0 cred proposal messages."""

    return (
        {
            "formats": [
                V20CredFormat(
                    attach_id=fmt_api,
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][fmt_api],
                )
                for fmt_api in filt_spec
            ],
            "filters_attach": [
                AttachDecorator.data_base64(filt_by_fmt, ident=fmt_api)
                for (fmt_api, filt_by_fmt) in filt_spec.items()
            ],
        }
        if filt_spec
        else {}
    )


async def _get_attached_credentials(
    profile: Profile, cred_ex_record: V20CredExRecord
) -> Mapping:
    """Fetch the detail records attached to a credential exchange."""
    result = {}

    for fmt in V20CredFormat.Format:
        detail_record = await fmt.handler(profile).get_detail_record(
            cred_ex_record.cred_ex_id
        )
        if detail_record:
            result[fmt.api] = detail_record

    return result


def _format_result_with_details(
    cred_ex_record: V20CredExRecord, details: Mapping
) -> Mapping:
    """Get credential exchange result with detail records."""
    result = {"cred_ex_record": cred_ex_record.serialize()}
    for fmt in V20CredFormat.Format:
        ident = fmt.api
        detail_record = details.get(ident)
        result[ident] = detail_record.serialize() if detail_record else None
    return result


@docs(
    tags=["issue-credential v2.0"],
    summary="Fetch all credential exchange records",
)
@querystring_schema(V20CredExRecordListQueryStringSchema)
@response_schema(V20CredExRecordListResultSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_list(request: web.BaseRequest):
    """Request handler for searching credential exchange records.

    Args:
        request: aiohttp request object

    Returns:
        The connection list response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    tag_filter = {}
    if "thread_id" in request.query and request.query["thread_id"] != "":
        tag_filter["thread_id"] = request.query["thread_id"]
    post_filter = {
        k: request.query[k]
        for k in ("connection_id", "role", "state")
        if request.query.get(k, "") != ""
    }

    limit, offset, order_by, descending = get_paginated_query_params(request)

    try:
        async with profile.session() as session:
            cred_ex_records = await V20CredExRecord.query(
                session=session,
                tag_filter=tag_filter,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
                post_filter_positive=post_filter,
            )

        results = []
        for cxr in cred_ex_records:
            details = await _get_attached_credentials(profile, cxr)
            result = _format_result_with_details(cxr, details)
            results.append(result)

    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(
    tags=["issue-credential v2.0"],
    summary="Fetch a single credential exchange record",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@response_schema(V20CredExRecordDetailSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_retrieve(request: web.BaseRequest):
    """Request handler for fetching single credential exchange record.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    cred_ex_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    try:
        async with profile.session() as session:
            cred_ex_record = await V20CredExRecord.retrieve_by_id(session, cred_ex_id)

        details = await _get_attached_credentials(profile, cred_ex_record)
        result = _format_result_with_details(cred_ex_record, details)
    except StorageNotFoundError as err:
        # no such cred ex record: not protocol error, user fat-fingered id
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        # present but broken or hopeless: protocol error
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary=(
        "Create a credential record without sending (generally for use with Out-Of-Band)"
    ),
)
@request_schema(V20IssueCredSchemaCore())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_create(request: web.BaseRequest):
    """Request handler for creating a credential from attr values.

    The internal credential record will be created without the credential
    being sent to any connection. This can be used in conjunction with
    the `oob` protocols to bind messages to an out of band message.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    body = await request.json()

    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    filt_spec = body.get("filter")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    if not filt_spec:
        raise web.HTTPBadRequest(reason="Missing filter")
    trace_msg = body.get("trace")

    try:
        # Not all formats use credential preview
        cred_preview = V20CredPreview.deserialize(preview_spec) if preview_spec else None
        cred_proposal = V20CredProposal(
            comment=comment,
            credential_preview=cred_preview,
            **_formats_filters(filt_spec),
        )
        cred_proposal.assign_trace_decorator(
            context.settings,
            trace_msg,
        )

        trace_event(
            context.settings,
            cred_proposal,
            outcome="credential_exchange_create.START",
        )

        cred_manager = V20CredManager(context.profile)
        (cred_ex_record, cred_offer_message) = await cred_manager.prepare_send(
            connection_id=None,
            cred_proposal=cred_proposal,
            auto_remove=auto_remove,
        )
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    trace_event(
        context.settings,
        cred_offer_message,
        outcome="credential_exchange_create.END",
        perf_counter=r_time,
    )

    return web.json_response(cred_ex_record.serialize())


@docs(
    tags=["issue-credential v2.0"],
    summary="Send holder a credential, automating entire flow",
)
@request_schema(V20CredExFreeSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send(request: web.BaseRequest):
    """Request handler for sending credential from issuer to holder from attr values.

    If both issuer and holder are configured for automatic responses, the operation
    ultimately results in credential issue; otherwise, the result waits on the first
    response not automated; the credential exchange record retains state regardless.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    comment = body.get("comment")
    connection_id = body.get("connection_id")
    verification_method = body.get("verification_method")

    filt_spec = body.get("filter")
    if not filt_spec:
        raise web.HTTPBadRequest(reason="Missing filter")
    preview_spec = body.get("credential_preview")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    replacement_id = body.get("replacement_id")
    trace_msg = body.get("trace")

    conn_record = None
    cred_ex_record = None
    try:
        # Not all formats use credential preview
        cred_preview = V20CredPreview.deserialize(preview_spec) if preview_spec else None
        async with profile.session() as session:
            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
        if not conn_record.is_ready:
            raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        # TODO: why do we create a proposal and then use that to create an offer.
        # Seems easier to just pass the proposal data to the format specific handler
        cred_proposal = V20CredProposal(
            comment=comment,
            credential_preview=cred_preview,
            **_formats_filters(filt_spec),
        )
        cred_proposal.assign_trace_decorator(
            context.settings,
            trace_msg,
        )

        trace_event(
            context.settings,
            cred_proposal,
            outcome="credential_exchange_send.START",
        )

        cred_manager = V20CredManager(profile)
        (cred_ex_record, cred_offer_message) = await cred_manager.prepare_send(
            connection_id,
            verification_method=verification_method,
            cred_proposal=cred_proposal,
            auto_remove=auto_remove,
            replacement_id=replacement_id,
        )
        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        LedgerError,
        StorageError,
        V20CredManagerError,
        V20CredFormatError,
    ) as err:
        # Only log full exception for unexpected errors
        if isinstance(err, (V20CredFormatError, V20CredManagerError)):
            LOGGER.warning(f"Error preparing credential offer: {err.roll_up}")
        else:
            LOGGER.exception("Error preparing credential offer")

        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record or conn_record,
            outbound_handler,
        )

    await outbound_handler(
        cred_offer_message,
        connection_id=cred_ex_record.connection_id,
    )

    trace_event(
        context.settings,
        cred_offer_message,
        outcome="credential_exchange_send.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Send issuer a credential proposal",
)
@request_schema(V20CredExFreeSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_proposal(request: web.BaseRequest):
    """Request handler for sending credential proposal.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    filt_spec = body.get("filter")
    if not filt_spec:
        raise web.HTTPBadRequest(reason="Missing filter")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    trace_msg = body.get("trace")

    conn_record = None
    cred_ex_record = None
    try:
        cred_preview = V20CredPreview.deserialize(preview_spec) if preview_spec else None
        async with profile.session() as session:
            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
        if not conn_record.is_ready:
            raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        cred_manager = V20CredManager(profile)
        cred_ex_record = await cred_manager.create_proposal(
            connection_id=connection_id,
            auto_remove=auto_remove,
            comment=comment,
            cred_preview=cred_preview,
            trace=trace_msg,
            fmt2filter={
                V20CredFormat.Format.get(fmt_api): filt_by_fmt
                for (fmt_api, filt_by_fmt) in filt_spec.items()
            },
        )

        cred_proposal_message = cred_ex_record.cred_proposal
        result = cred_ex_record.serialize()

    except (BaseModelError, StorageError) as err:
        LOGGER.exception("Error preparing credential proposal")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cannot yet receive a problem report about our failed protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(cred_proposal_message, connection_id=connection_id)

    trace_event(
        context.settings,
        cred_proposal_message,
        outcome="credential_exchange_send_proposal.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


async def _create_free_offer(
    profile: Profile,
    filt_spec: Optional[Mapping] = None,
    connection_id: Optional[str] = None,
    auto_issue: bool = False,
    auto_remove: bool = False,
    replacement_id: Optional[str] = None,
    preview_spec: Optional[dict] = None,
    comment: Optional[str] = None,
    trace_msg: Optional[bool] = None,
):
    """Create a credential offer and related exchange record."""

    cred_preview = V20CredPreview.deserialize(preview_spec) if preview_spec else None
    cred_proposal = V20CredProposal(
        comment=comment,
        credential_preview=cred_preview,
        **_formats_filters(filt_spec),
    )
    cred_proposal.assign_trace_decorator(
        profile.settings,
        trace_msg,
    )

    cred_ex_record = V20CredExRecord(
        connection_id=connection_id,
        initiator=V20CredExRecord.INITIATOR_SELF,
        role=V20CredExRecord.ROLE_ISSUER,
        cred_proposal=cred_proposal.serialize(),
        auto_issue=auto_issue,
        auto_remove=auto_remove,
        trace=trace_msg,
    )

    cred_manager = V20CredManager(profile)
    try:
        (cred_ex_record, cred_offer_message) = await cred_manager.create_offer(
            cred_ex_record,
            comment=comment,
            replacement_id=replacement_id,
        )
    except ValueError as err:
        LOGGER.exception(f"Error creating credential offer: {err}")
        async with profile.session() as session:
            await cred_ex_record.save_error_state(session, reason=err)
        raise web.HTTPBadRequest(reason=err)

    return (cred_ex_record, cred_offer_message)


@docs(
    tags=["issue-credential v2.0"],
    summary="Create a credential offer, independent of any proposal or connection",
)
@request_schema(V20CredOfferConnFreeRequestSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_create_free_offer(request: web.BaseRequest):
    """Request handler for creating free credential offer.

    Unlike with `send-offer`, this credential exchange is not tied to a specific
    connection. It must be dispatched out-of-band by the controller.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile

    body = await request.json()

    auto_issue = body.get(
        "auto_issue", context.settings.get("debug.auto_respond_credential_request")
    )
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    replacement_id = body.get("replacement_id")
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    filt_spec = body.get("filter")
    if not filt_spec:
        raise web.HTTPBadRequest(reason="Missing filter")
    trace_msg = body.get("trace")
    cred_ex_record = None
    try:
        (cred_ex_record, cred_offer_message) = await _create_free_offer(
            profile=profile,
            filt_spec=filt_spec,
            auto_issue=auto_issue,
            auto_remove=auto_remove,
            replacement_id=replacement_id,
            preview_spec=preview_spec,
            comment=comment,
            trace_msg=trace_msg,
        )
        result = cred_ex_record.serialize()
    except (
        BaseModelError,
        LedgerError,
        V20CredFormatError,
        V20CredManagerError,
    ) as err:
        LOGGER.exception("Error creating free credential offer")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        raise web.HTTPBadRequest(reason=err.roll_up)
    trace_event(
        context.settings,
        cred_offer_message,
        outcome="credential_exchange_create_free_offer.END",
        perf_counter=r_time,
    )
    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Send holder a credential offer, independent of any proposal",
)
@request_schema(V20CredOfferRequestSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_free_offer(request: web.BaseRequest):
    """Request handler for sending free credential offer.

    An issuer initiates a such a credential offer, free from any
    holder-initiated corresponding credential proposal with preview.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    filt_spec = body.get("filter")
    if not filt_spec:
        raise web.HTTPBadRequest(reason="Missing filter")
    auto_issue = body.get(
        "auto_issue", context.settings.get("debug.auto_respond_credential_request")
    )
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    replacement_id = body.get("replacement_id")
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    trace_msg = body.get("trace")

    cred_ex_record = None
    conn_record = None
    try:
        async with profile.session() as session:
            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
        if not conn_record.is_ready:
            raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        cred_ex_record, cred_offer_message = await _create_free_offer(
            profile=profile,
            filt_spec=filt_spec,
            connection_id=connection_id,
            auto_issue=auto_issue,
            auto_remove=auto_remove,
            preview_spec=preview_spec,
            comment=comment,
            trace_msg=trace_msg,
            replacement_id=replacement_id,
        )
        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        AnonCredsIssuerError,
        IndyIssuerError,
        LedgerError,
        StorageNotFoundError,
        V20CredFormatError,
        V20CredManagerError,
    ) as err:
        LOGGER.exception("Error preparing free credential offer")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cannot yet receive a problem report about our failed protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(cred_offer_message, connection_id=connection_id)

    trace_event(
        context.settings,
        cred_offer_message,
        outcome="credential_exchange_send_free_offer.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Send holder a credential offer in reference to a proposal with preview",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@request_schema(V20CredBoundOfferRequestSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_bound_offer(request: web.BaseRequest):
    """Request handler for sending bound credential offer.

    A holder initiates this sequence with a credential proposal; this message
    responds with an offer bound to the proposal.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json() if request.body_exists else {}
    filt_spec = body.get("filter")
    preview_spec = body.get("counter_preview")

    cred_ex_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    conn_record = None
    try:
        async with profile.session() as session:
            try:
                cred_ex_record = await V20CredExRecord.retrieve_by_id(
                    session,
                    cred_ex_id,
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            connection_id = cred_ex_record.connection_id
            if cred_ex_record.state != (
                V20CredExRecord.STATE_PROPOSAL_RECEIVED
            ):  # check state here: manager call creates free offers too
                raise V20CredManagerError(
                    f"Credential exchange record {cred_ex_record.cred_ex_id} "
                    f"in {cred_ex_record.state} state "
                    f"(must be {V20CredExRecord.STATE_PROPOSAL_RECEIVED})"
                )

            conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not conn_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        cred_manager = V20CredManager(profile)
        (cred_ex_record, cred_offer_message) = await cred_manager.create_offer(
            cred_ex_record,
            counter_proposal=(
                V20CredProposal(
                    comment=None,
                    credential_preview=V20CredPreview.deserialize(preview_spec),
                    **_formats_filters(filt_spec),
                )
                if preview_spec
                else None
            ),
            comment=None,
        )

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        AnonCredsIssuerError,
        IndyIssuerError,
        LedgerError,
        StorageError,
        V20CredFormatError,
        V20CredManagerError,
    ) as err:
        LOGGER.exception("Error preparing bound credential offer")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )
    except LinkedDataProofException as err:
        raise web.HTTPBadRequest(reason=str(err)) from err

    await outbound_handler(cred_offer_message, connection_id=connection_id)

    trace_event(
        context.settings,
        cred_offer_message,
        outcome="credential_exchange_send_bound_offer.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary=(
        "Send issuer a credential request not bound to an existing thread."
        " Indy credentials cannot start at a request"
    ),
)
@request_schema(V20CredRequestFreeSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_free_request(request: web.BaseRequest):
    """Request handler for sending free credential request.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    comment = body.get("comment")
    filt_spec = body.get("filter")
    if not filt_spec:
        raise web.HTTPBadRequest(reason="Missing filter")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    trace_msg = body.get("trace")
    holder_did = body.get("holder_did")

    conn_record = None
    cred_ex_record = None
    try:
        try:
            async with profile.session() as session:
                conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not conn_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")
        except StorageNotFoundError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        cred_manager = V20CredManager(profile)

        cred_proposal = V20CredProposal(
            comment=comment,
            **_formats_filters(filt_spec),
        )

        cred_ex_record = V20CredExRecord(
            connection_id=connection_id,
            auto_remove=auto_remove,
            cred_proposal=cred_proposal.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_HOLDER,
            trace=trace_msg,
        )

        cred_ex_record, cred_request_message = await cred_manager.create_request(
            cred_ex_record=cred_ex_record,
            holder_did=holder_did,
            comment=comment,
        )

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        AnonCredsHolderError,
        IndyHolderError,
        LedgerError,
        StorageError,
        V20CredManagerError,
    ) as err:
        LOGGER.exception("Error preparing free credential request")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cannot yet receive a problem report about our failed protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(cred_request_message, connection_id=connection_id)

    trace_event(
        context.settings,
        cred_request_message,
        outcome="credential_exchange_send_free_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Send issuer a credential request",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@request_schema(V20CredRequestRequestSchema())
@response_schema(V20CredExRecordSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_bound_request(request: web.BaseRequest):
    """Request handler for sending credential request.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    try:
        body = await request.json() or {}
        holder_did = body.get("holder_did")
        auto_remove = body.get(
            "auto_remove", not profile.settings.get("preserve_exchange_records")
        )
    except JSONDecodeError:
        holder_did = None
        auto_remove = not profile.settings.get("preserve_exchange_records")

    cred_ex_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    conn_record = None
    try:
        async with profile.session() as session:
            try:
                cred_ex_record = await V20CredExRecord.retrieve_by_id(
                    session,
                    cred_ex_id,
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            conn_record = None
            if cred_ex_record.connection_id:
                try:
                    conn_record = await ConnRecord.retrieve_by_id(
                        session, cred_ex_record.connection_id
                    )
                except StorageNotFoundError as err:
                    raise web.HTTPBadRequest(reason=err.roll_up) from err

        if conn_record and not conn_record.is_ready:
            raise web.HTTPForbidden(
                reason=f"Connection {cred_ex_record.connection_id} not ready"
            )

        if conn_record or holder_did:
            holder_did = holder_did or conn_record.my_did
        else:
            # Need to get the holder DID from the out of band record
            async with profile.session() as session:
                oob_record = await OobRecord.retrieve_by_tag_filter(
                    session,
                    {"invi_msg_id": cred_ex_record.cred_offer._thread.pthid},
                )
                # Transform recipient key into did
                holder_did = default_did_from_verkey(oob_record.our_recipient_key)

        # assign the auto_remove flag from above...
        cred_ex_record.auto_remove = auto_remove

        cred_manager = V20CredManager(profile)
        cred_ex_record, cred_request_message = await cred_manager.create_request(
            cred_ex_record,
            holder_did,
        )

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        AnonCredsHolderError,
        IndyHolderError,
        LedgerError,
        StorageError,
        V20CredFormatError,
        V20CredManagerError,
    ) as err:
        LOGGER.exception("Error preparing bound credential request")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    await outbound_handler(
        cred_request_message, connection_id=cred_ex_record.connection_id
    )

    trace_event(
        context.settings,
        cred_request_message,
        outcome="credential_exchange_send_bound_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Send holder a credential",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@request_schema(V20CredIssueRequestSchema())
@response_schema(V20CredExRecordDetailSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_issue(request: web.BaseRequest):
    """Request handler for sending credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()
    comment = body.get("comment")

    cred_ex_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    conn_record = None
    try:
        async with profile.session() as session:
            try:
                cred_ex_record = await V20CredExRecord.retrieve_by_id(
                    session,
                    cred_ex_id,
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            conn_record = None
            if cred_ex_record.connection_id:
                conn_record = await ConnRecord.retrieve_by_id(
                    session, cred_ex_record.connection_id
                )
        if conn_record and not conn_record.is_ready:
            raise web.HTTPForbidden(
                reason=f"Connection {cred_ex_record.connection_id} not ready"
            )

        cred_manager = V20CredManager(profile)
        (cred_ex_record, cred_issue_message) = await cred_manager.issue_credential(
            cred_ex_record,
            comment=comment,
        )

        details = await _get_attached_credentials(profile, cred_ex_record)
        result = _format_result_with_details(cred_ex_record, details)

    except (
        BaseModelError,
        AnonCredsIssuerError,
        IndyIssuerError,
        LedgerError,
        StorageError,
        V20CredFormatError,
        V20CredManagerError,
    ) as err:
        LOGGER.exception("Error preparing issued credential")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    await outbound_handler(cred_issue_message, connection_id=cred_ex_record.connection_id)

    trace_event(
        context.settings,
        cred_issue_message,
        outcome="credential_exchange_issue.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Store a received credential",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@request_schema(V20CredStoreRequestSchema())
@response_schema(V20CredExRecordDetailSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_store(request: web.BaseRequest):
    """Request handler for storing credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    try:
        body = await request.json() or {}
        cred_id = body.get("credential_id")
    except JSONDecodeError:
        cred_id = None

    cred_ex_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    conn_record = None
    try:
        async with profile.session() as session:
            try:
                cred_ex_record = await V20CredExRecord.retrieve_by_id(
                    session,
                    cred_ex_id,
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            conn_record = None
            if cred_ex_record.connection_id:
                conn_record = await ConnRecord.retrieve_by_id(
                    session, cred_ex_record.connection_id
                )
            if conn_record and not conn_record.is_ready:
                raise web.HTTPForbidden(
                    reason=f"Connection {cred_ex_record.connection_id} not ready"
                )

        cred_manager = V20CredManager(profile)
        cred_ex_record = await cred_manager.store_credential(cred_ex_record, cred_id)

    except (
        AnonCredsHolderError,
        IndyHolderError,
        StorageError,
        V20CredManagerError,
    ) as err:  # treat failure to store as mangled on receipt hence protocol error
        LOGGER.exception("Error storing issued credential")
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    try:
        # fetch these early, before potential removal
        details = await _get_attached_credentials(profile, cred_ex_record)

        # the record may be auto-removed here
        (
            cred_ex_record,
            cred_ack_message,
        ) = await cred_manager.send_cred_ack(cred_ex_record)

        result = _format_result_with_details(cred_ex_record, details)

    except (
        BaseModelError,
        StorageError,
        V20CredFormatError,
        V20CredManagerError,
    ) as err:
        # protocol finished OK: do not send problem report nor set record state error
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    trace_event(
        context.settings,
        cred_ack_message,
        outcome="credential_exchange_store.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v2.0"],
    summary="Remove an existing credential exchange record",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@response_schema(V20IssueCredentialModuleResponseSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_remove(request: web.BaseRequest):
    """Request handler for removing a credential exchange record.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]

    cred_ex_id = request.match_info["cred_ex_id"]
    try:
        cred_manager = V20CredManager(context.profile)
        await cred_manager.delete_cred_ex_record(cred_ex_id)
    except StorageNotFoundError as err:  # not a protocol error
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:  # not a protocol error
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


@docs(
    tags=["issue-credential v2.0"],
    summary="Send a problem report for credential exchange",
)
@match_info_schema(V20CredExIdMatchInfoSchema())
@request_schema(V20CredIssueProblemReportRequestSchema())
@response_schema(V20IssueCredentialModuleResponseSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_problem_report(request: web.BaseRequest):
    """Request handler for sending problem report.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    cred_ex_id = request.match_info["cred_ex_id"]
    body = await request.json()
    description = body["description"]

    try:
        async with profile.session() as session:
            cred_ex_record = await V20CredExRecord.retrieve_by_id(session, cred_ex_id)
            report = problem_report_for_record(cred_ex_record, description)
            await cred_ex_record.save_error_state(
                session,
                reason=f"created problem report: {description}",
            )
    except StorageNotFoundError as err:  # other party does not care about meta-problems
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(report, connection_id=cred_ex_record.connection_id)

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get(
                "/issue-credential-2.0/records",
                credential_exchange_list,
                allow_head=False,
            ),
            web.post(
                "/issue-credential-2.0/create-offer",
                credential_exchange_create_free_offer,
            ),
            web.get(
                "/issue-credential-2.0/records/{cred_ex_id}",
                credential_exchange_retrieve,
                allow_head=False,
            ),
            web.post("/issue-credential-2.0/create", credential_exchange_create),
            web.post("/issue-credential-2.0/send", credential_exchange_send),
            web.post(
                "/issue-credential-2.0/send-proposal", credential_exchange_send_proposal
            ),
            web.post(
                "/issue-credential-2.0/send-offer", credential_exchange_send_free_offer
            ),
            web.post(
                "/issue-credential-2.0/send-request",
                credential_exchange_send_free_request,
            ),
            web.post(
                "/issue-credential-2.0/records/{cred_ex_id}/send-offer",
                credential_exchange_send_bound_offer,
            ),
            web.post(
                "/issue-credential-2.0/records/{cred_ex_id}/send-request",
                credential_exchange_send_bound_request,
            ),
            web.post(
                "/issue-credential-2.0/records/{cred_ex_id}/issue",
                credential_exchange_issue,
            ),
            web.post(
                "/issue-credential-2.0/records/{cred_ex_id}/store",
                credential_exchange_store,
            ),
            web.post(
                "/issue-credential-2.0/records/{cred_ex_id}/problem-report",
                credential_exchange_problem_report,
            ),
            web.delete(
                "/issue-credential-2.0/records/{cred_ex_id}",
                credential_exchange_remove,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "issue-credential v2.0",
            "description": "Credential issue v2.0",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



BBS_PROOF_VC_MATTR = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/citizenship/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
    "type": ["PermanentResidentCard", "VerifiableCredential"],
    "description": "Government of Example Permanent Resident Card.",
    "identifier": "83627465",
    "name": "Permanent Resident Card",
    "credentialSubject": {
        "id": "did:example:b34ca6cd37bbf23",
        "type": ["Person", "PermanentResident"],
        "familyName": "SMITH",
        "gender": "Male",
        "givenName": "JOHN",
    },
    "expirationDate": "2029-12-03T12:19:52Z",
    "issuanceDate": "2019-12-03T12:19:52Z",
    "issuer": "did:example:489398593",
    "proof": {
        "type": "BbsBlsSignatureProof2020",
        "nonce": "wrmPiSRm+iBqnGBXz+/37LLYRZWirGgIORKHIkrgWVnHtb4fDe/4ZPZaZ+/RwGVJYYY=",
        "proofValue": "ABkB/wbvt6213E9eJ+aRGbdG1IIQtx+IdAXALLNg2a5ENSGOIBxRGSoArKXwD/diieDWG6+0q8CWh7CViUqOOdEhYp/DonzmjoWbWECalE6x/qtyBeE7W9TJTXyK/yW6JKSKPz2ht4J0XLV84DZrxMF4HMrY7rFHvdE4xV7ULeC9vNmAmwYAqJfNwY94FG2erg2K2cg0AAAAdLfutjMuBO0JnrlRW6O6TheATv0xZZHP9kf1AYqPaxsYg0bq2XYzkp+tzMBq1rH3tgAAAAIDTzuPazvFHijdzuAgYg+Sg0ziF+Gw5Bz8r2cuvuSg1yKWqW1dM5GhGn6SZUpczTXuZuKGlo4cZrwbIg9wf4lBs3kQwWULRtQUXki9izmznt4Go98X/ElOguLLum4S78Gehe1ql6CXD1zS5PiDXjDzAAAACWz/sbigWpPmUqNA8YUczOuzBUvzmkpjVyL9aqf1e7rSZmN8CNa6dTGOzgKYgDGoIbSQR8EN8Ld7kpTIAdi4YvNZwEYlda/BR6oSrFCquafz7s/jeXyOYMsiVC53Zls9KEg64tG7n90XuZOyMk9RAdcxYRGligbFuG2Ap+rQ+rrELJaW7DWwFEI6cRnitZo6aS0hHmiOKKtJyA7KFbx27nBGd2y3JCvgYO6VUROQ//t3F4aRVI1U53e5N3MU+lt9GmFeL+Kv+2zV1WssScO0ZImDGDOvjDs1shnNSjIJ0RBNAo2YzhFKh3ExWd9WbiZ2/USSyomaSK4EzdTDqi2JCGdqS7IpooKSX/1Dp4K+d8HhPLGNLX4yfMoG9SnRfRQZZQ==",
        "verificationMethod": "did:example:489398593#test",
        "proofPurpose": "assertionMethod",
        "created": "2020-10-16T23:59:31Z",
    },
}



"""Manager for multiple ledger."""

from abc import ABC, abstractmethod
from typing import List, Mapping, Optional, Tuple

from ...core.error import BaseError
from ...core.profile import Profile
from ...ledger.base import BaseLedger
from ...messaging.valid import IndyDID
from ...multitenant.manager import BaseMultitenantManager


class MultipleLedgerManagerError(BaseError):
    """Generic multiledger error."""


class BaseMultipleLedgerManager(ABC):
    """Base class for handling multiple ledger support."""

    def __init__(self, profile: Profile):
        """Initialize Multiple Ledger Manager."""

    @abstractmethod
    def get_endorser_info_for_ledger(self, ledger_id: str) -> Optional[Tuple[str, str]]:
        """Return endorser alias, did tuple for provided ledger, if available."""

    @abstractmethod
    async def get_write_ledgers(self) -> List[str]:
        """Return write ledger."""

    @abstractmethod
    async def get_ledger_id_by_ledger_pool_name(self, pool_name: str) -> str:
        """Return ledger_id by ledger pool name."""

    @abstractmethod
    async def get_prod_ledgers(self) -> Mapping:
        """Return configured production ledgers."""

    @abstractmethod
    async def get_nonprod_ledgers(self) -> Mapping:
        """Return configured non production ledgers."""

    @abstractmethod
    async def _get_ledger_by_did(
        self, ledger_id: str, did: str
    ) -> Optional[Tuple[str, BaseLedger, bool]]:
        """Build and submit GET_NYM request and process response."""

    @abstractmethod
    async def get_ledger_inst_by_id(self, ledger_id: str) -> Optional[BaseLedger]:
        """Return ledger instance by identifier."""

    @abstractmethod
    async def lookup_did_in_configured_ledgers(
        self, did: str, cache_did: bool
    ) -> Tuple[str, BaseLedger]:
        """Lookup given DID in configured ledgers in parallel."""

    def extract_did_from_identifier(self, identifier: str) -> str:
        """Return did from record identifier (REV_REG_ID, CRED_DEF_ID, SCHEMA_ID)."""
        if bool(IndyDID.PATTERN.match(identifier)):
            return identifier.split(":")[-1]
        else:
            return identifier.split(":")[0]

    async def set_profile_write_ledger(self, ledger_id: str, profile: Profile) -> str:
        """Set the write ledger for the profile."""
        if ledger_id not in self.writable_ledgers:
            raise MultipleLedgerManagerError(
                f"Provided Ledger identifier {ledger_id} is not write configurable."
            )
        extra_settings = {}
        multi_tenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        multi_ledgers = self.production_ledgers | self.non_production_ledgers
        if ledger_id in multi_ledgers:
            profile.context.injector.bind_instance(
                BaseLedger, multi_ledgers.get(ledger_id)
            )
            self._update_settings(profile.context.settings, ledger_id)
            self._update_settings(extra_settings, ledger_id)
            if multi_tenant_mgr:
                await multi_tenant_mgr.update_wallet(
                    profile.context.settings["wallet.id"],
                    extra_settings,
                )
            return ledger_id
        raise MultipleLedgerManagerError(f"No ledger info found for {ledger_id}.")

    def _update_settings(self, settings, ledger_id: str):
        endorser_info = self.get_endorser_info_for_ledger(ledger_id)
        if endorser_info:
            endorser_alias, endorser_did = endorser_info
            settings["endorser.endorser_alias"] = endorser_alias
            settings["endorser.endorser_public_did"] = endorser_did
        settings["ledger.write_ledger"] = ledger_id



import pytest
from marshmallow import ValidationError

from acapy_agent.storage.base import DEFAULT_PAGE_SIZE, MAXIMUM_PAGE_SIZE

from ..paginated_query import PaginatedQuerySchema


def test_paginated_query_schema_defaults():
    schema = PaginatedQuerySchema()
    result = schema.load({})
    assert result["limit"] == DEFAULT_PAGE_SIZE
    assert result["offset"] == 0


def test_paginated_query_schema_not_required():
    schema = PaginatedQuerySchema()
    result = schema.load({"limit": 10})
    assert result["limit"] == 10
    assert result["offset"] == 0

    result = schema.load({"offset": 5})
    assert result["limit"] == DEFAULT_PAGE_SIZE
    assert result["offset"] == 5


def test_paginated_query_schema_limit_validation():
    schema = PaginatedQuerySchema()

    # Valid limit
    result = schema.load({"limit": 1})
    assert result["limit"] == 1

    result = schema.load({"limit": MAXIMUM_PAGE_SIZE})
    assert result["limit"] == MAXIMUM_PAGE_SIZE

    # Invalid limit (less than 1)
    with pytest.raises(ValidationError) as exc_info:
        schema.load({"limit": 0})
    assert (
        f"Must be greater than or equal to 1 and less than or equal to {MAXIMUM_PAGE_SIZE}"
        in str(exc_info.value)
    )

    # Invalid limit (greater than MAXIMUM_PAGE_SIZE)
    with pytest.raises(ValidationError) as exc_info:
        schema.load({"limit": MAXIMUM_PAGE_SIZE + 1})
    assert (
        f"Must be greater than or equal to 1 and less than or equal to {MAXIMUM_PAGE_SIZE}"
        in str(exc_info.value)
    )


def test_paginated_query_schema_offset_validation():
    schema = PaginatedQuerySchema()

    # Valid offset
    result = schema.load({"offset": 0})
    assert result["offset"] == 0

    result = schema.load({"offset": 10})
    assert result["offset"] == 10

    # Invalid offset (less than 0)
    with pytest.raises(ValidationError) as exc_info:
        schema.load({"offset": -1})
    assert "Must be greater than or equal to 0." in str(exc_info.value)



import json
from unittest import TestCase

from ..jwe import JweEnvelope, JweRecipient, b64url, from_b64url

IV = b"test nonce"
TAG = b"test tag"
AAD = b"test aad"
CIPHERTEXT = b"test ciphertext"
ENC_KEY_1 = b"test enc key 1"
ENC_KEY_2 = b"test enc key 2"
PARAMS = {"alg": "MyAlg"}
UNPROTECTED = {"abc": "ABC"}


class TestJwe(TestCase):
    def test_envelope_load_single_recipient(self):
        protected = PARAMS.copy()
        protected.update(
            {
                # flattened single recipient
                "header": {"def": "DEF"},
                "encrypted_key": b64url(ENC_KEY_1),
            }
        )
        message = {
            "protected": b64url(json.dumps(protected)),
            "unprotected": UNPROTECTED.copy(),
            "iv": b64url(IV),
            "ciphertext": b64url(CIPHERTEXT),
            "tag": b64url(TAG),
            "aad": b64url(AAD),
        }
        loaded = JweEnvelope.deserialize(message)

        assert loaded.protected == PARAMS
        assert loaded.unprotected == UNPROTECTED
        assert loaded.iv == IV
        assert loaded.tag == TAG
        assert loaded.aad == AAD
        assert loaded.combined_aad == loaded.protected_bytes + b"." + b64url(AAD).encode(
            "utf-8"
        )
        assert loaded.ciphertext == CIPHERTEXT

        recips = list(loaded.recipients)
        assert len(recips) == 1
        assert recips[0].encrypted_key == ENC_KEY_1
        assert recips[0].header == {"alg": "MyAlg", "abc": "ABC", "def": "DEF"}

    def test_envelope_load_multiple_recipients(self):
        protected = PARAMS.copy()
        protected.update(
            {
                "recipients": [
                    {"header": {"def": "DEF"}, "encrypted_key": b64url(ENC_KEY_1)},
                    {"header": {"ghi": "GHI"}, "encrypted_key": b64url(ENC_KEY_2)},
                ]
            }
        )
        message = {
            "protected": b64url(json.dumps(protected)),
            "unprotected": UNPROTECTED.copy(),
            "iv": b64url(IV),
            "ciphertext": b64url(CIPHERTEXT),
            "tag": b64url(TAG),
            "aad": b64url(AAD),
        }
        loaded = JweEnvelope.deserialize(message)

        assert loaded.protected == PARAMS
        assert loaded.unprotected == UNPROTECTED
        assert loaded.iv == IV
        assert loaded.tag == TAG
        assert loaded.aad == AAD
        assert loaded.combined_aad == loaded.protected_bytes + b"." + b64url(AAD).encode(
            "utf-8"
        )
        assert loaded.ciphertext == CIPHERTEXT

        recips = list(loaded.recipients)
        assert len(recips) == 2
        assert recips[0].encrypted_key == ENC_KEY_1
        assert recips[0].header == {"alg": "MyAlg", "abc": "ABC", "def": "DEF"}
        assert recips[1].encrypted_key == ENC_KEY_2
        assert recips[1].header == {"alg": "MyAlg", "abc": "ABC", "ghi": "GHI"}

    def test_envelope_serialize_single_recipient(self):
        env = JweEnvelope(
            unprotected=UNPROTECTED.copy(),
            iv=IV,
            ciphertext=CIPHERTEXT,
            tag=TAG,
            aad=AAD,
            with_protected_recipients=True,
            with_flatten_recipients=True,
        )
        env.add_recipient(JweRecipient(encrypted_key=ENC_KEY_1, header={"def": "DEF"}))
        env.set_protected(PARAMS)
        message = env.to_json()
        loaded = JweEnvelope.from_json(message)

        # check in flattened form
        prot = json.loads(from_b64url(loaded.protected_b64))
        assert "encrypted_key" in prot

        assert loaded.protected == PARAMS
        assert loaded.with_protected_recipients
        assert loaded.with_flatten_recipients
        assert loaded.unprotected == UNPROTECTED
        assert loaded.iv == IV
        assert loaded.tag == TAG
        assert loaded.aad == AAD
        assert loaded.combined_aad == loaded.protected_bytes + b"." + b64url(AAD).encode(
            "utf-8"
        )
        assert loaded.ciphertext == CIPHERTEXT

        recips = list(loaded.recipients)
        assert len(recips) == 1
        assert recips[0].encrypted_key == ENC_KEY_1
        assert recips[0].header == {"alg": "MyAlg", "abc": "ABC", "def": "DEF"}

    def test_envelope_serialize_multiple_recipients(self):
        env = JweEnvelope(
            unprotected=UNPROTECTED.copy(),
            iv=IV,
            ciphertext=CIPHERTEXT,
            tag=TAG,
            aad=AAD,
            with_protected_recipients=True,
            with_flatten_recipients=True,
        )
        env.add_recipient(JweRecipient(encrypted_key=ENC_KEY_1, header={"def": "DEF"}))
        env.add_recipient(JweRecipient(encrypted_key=ENC_KEY_2, header={"ghi": "GHI"}))
        env.set_protected(PARAMS)
        message = env.to_json()
        loaded = JweEnvelope.from_json(message)

        assert loaded.protected == PARAMS
        assert loaded.with_protected_recipients
        assert not loaded.with_flatten_recipients
        assert loaded.unprotected == UNPROTECTED
        assert loaded.iv == IV
        assert loaded.tag == TAG
        assert loaded.aad == AAD
        assert loaded.combined_aad == loaded.protected_bytes + b"." + b64url(AAD).encode(
            "utf-8"
        )
        assert loaded.ciphertext == CIPHERTEXT

        recips = list(loaded.recipients)
        assert len(recips) == 2
        assert recips[0].encrypted_key == ENC_KEY_1
        assert recips[0].header == {"alg": "MyAlg", "abc": "ABC", "def": "DEF"}
        assert recips[1].encrypted_key == ENC_KEY_2
        assert recips[1].header == {"alg": "MyAlg", "abc": "ABC", "ghi": "GHI"}



"""Handle transaction information interface."""

from typing import Optional

from marshmallow import fields

from .....core.profile import ProfileSession
from .....messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from .....messaging.valid import UUID4_EXAMPLE
from ..controller import (
    ENDORSE_TRANSACTION,
    REFUSE_TRANSACTION,
    REGISTER_PUBLIC_DID,
    WRITE_DID_TRANSACTION,
    WRITE_TRANSACTION,
)


class TransactionRecord(BaseExchangeRecord):
    """Represents a single transaction record."""

    class Meta:
        """Transaction Record metadata."""

        schema_class = "TransactionRecordSchema"

    RECORD_ID_NAME = "transaction_id"
    CACHE_ENABLED = True
    TAG_NAMES = {"state", "thread_id", "connection_id"}
    RECORD_TYPE = "transaction"
    STATE_INIT = "init"
    RECORD_TOPIC = "endorse_transaction"

    SIGNATURE_REQUEST = "https://didcomm.org/sign-attachment/1.0/signature-request"

    SIGNATURE_RESPONSE = "https://didcomm.org/sign-attachment/1.0/signature-response"

    SIGNATURE_TYPE = "default"

    SIGNATURE_CONTEXT = "did:sov"

    ADD_SIGNATURE = "add-signature"

    ENDORSE_TRANSACTION = ENDORSE_TRANSACTION
    REFUSE_TRANSACTION = REFUSE_TRANSACTION
    WRITE_TRANSACTION = WRITE_TRANSACTION
    WRITE_DID_TRANSACTION = WRITE_DID_TRANSACTION
    REGISTER_PUBLIC_DID = REGISTER_PUBLIC_DID

    FORMAT_VERSION = "dif/endorse-transaction/request@v1.0"

    STATE_TRANSACTION_CREATED = "transaction_created"
    STATE_REQUEST_SENT = "request_sent"
    STATE_REQUEST_RECEIVED = "request_received"
    STATE_TRANSACTION_ENDORSED = "transaction_endorsed"
    STATE_TRANSACTION_REFUSED = "transaction_refused"
    STATE_TRANSACTION_RESENT = "transaction_resent"
    STATE_TRANSACTION_RESENT_RECEIVED = "transaction_resent_received"
    STATE_TRANSACTION_CANCELLED = "transaction_cancelled"
    STATE_TRANSACTION_ACKED = "transaction_acked"

    def __init__(
        self,
        *,
        transaction_id: Optional[str] = None,
        _type: Optional[str] = None,
        comment: Optional[str] = None,
        signature_request: Optional[list] = None,
        signature_response: Optional[list] = None,
        timing: Optional[dict] = None,
        formats: Optional[list] = None,
        messages_attach: Optional[list] = None,
        thread_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        state: Optional[str] = None,
        endorser_write_txn: Optional[bool] = None,
        meta_data: Optional[dict] = None,
        **kwargs,
    ):
        """Initialize a new TransactionRecord."""

        super().__init__(transaction_id, state or self.STATE_INIT, **kwargs)
        self._type = _type
        self.comment = comment
        self.signature_request = signature_request or []
        self.signature_response = signature_response or []
        self.timing = timing or {}
        self.formats = formats or []
        self.messages_attach = messages_attach or []
        self.thread_id = thread_id
        self.connection_id = connection_id
        self.endorser_write_txn = endorser_write_txn
        self.meta_data = meta_data or {"context": {}, "processing": {}}

    @property
    def transaction_id(self) -> str:
        """Accessor for the ID associated with this record."""
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this transaction record."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "comment",
                "signature_request",
                "signature_response",
                "timing",
                "formats",
                "messages_attach",
                "thread_id",
                "connection_id",
                "state",
                "endorser_write_txn",
                "meta_data",
            )
        }

    @classmethod
    async def retrieve_by_connection_and_thread(
        cls, session: ProfileSession, connection_id: str, thread_id: str
    ) -> "TransactionRecord":
        """Retrieve a TransactionRecord by connection and thread ID."""
        cache_key = f"transaction_record_ctidx::{connection_id}::{thread_id}"
        record_id = await cls.get_cached_key(session, cache_key)
        if record_id:
            record = await cls.retrieve_by_id(session, record_id)
        else:
            record = await cls.retrieve_by_tag_filter(
                session,
                {"thread_id": thread_id},
                {"connection_id": connection_id} if connection_id else None,
            )
            await cls.set_cached_key(session, cache_key, record._id)
        return record


class TransactionRecordSchema(BaseExchangeSchema):
    """Schema to allow serialization/deserialization of transaction records."""

    class Meta:
        """TransactionRecordSchema metadata."""

        model_class = "TransactionRecord"

    transaction_id = fields.Str(
        required=False,
        metadata={"description": "Transaction identifier", "example": UUID4_EXAMPLE},
    )
    _type = fields.Str(
        required=False, metadata={"description": "Transaction type", "example": "101"}
    )
    signature_request = fields.List(
        fields.Dict(
            metadata={
                "example": {
                    "context": TransactionRecord.SIGNATURE_CONTEXT,
                    "method": TransactionRecord.ADD_SIGNATURE,
                    "signature_type": TransactionRecord.SIGNATURE_TYPE,
                    "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
                    "author_goal_code": TransactionRecord.WRITE_TRANSACTION,
                }
            }
        ),
        required=False,
    )
    signature_response = fields.List(
        fields.Dict(
            metadata={
                "example": {
                    "message_id": UUID4_EXAMPLE,
                    "context": TransactionRecord.SIGNATURE_CONTEXT,
                    "method": TransactionRecord.ADD_SIGNATURE,
                    "signer_goal_code": TransactionRecord.REFUSE_TRANSACTION,
                }
            }
        ),
        required=False,
    )
    timing = fields.Dict(
        required=False,
        metadata={"example": {"expires_time": "2020-12-13T17:29:06+0000"}},
    )
    formats = fields.List(
        fields.Dict(
            keys=fields.Str(),
            values=fields.Str(),
            metadata={
                "example": {
                    "attach_id": UUID4_EXAMPLE,
                    "format": TransactionRecord.FORMAT_VERSION,
                }
            },
        ),
        required=False,
    )
    messages_attach = fields.List(
        fields.Dict(
            metadata={
                "example": {
                    "@id": "143c458d-1b1c-40c7-ab85-4d16808ddf0a",
                    "mime-type": "application/json",
                    "data": {
                        "json": (
                            '{"endorser": "V4SGRU86Z58d6TV7PBUe6f","identifier":'
                            ' "LjgpST2rjsoxYegQDRm7EL","operation": {"data":'
                            ' {"attr_names": ["first_name", "last_name"],"name":'
                            ' "test_schema","version": "2.1",},"type":'
                            ' "101",},"protocolVersion": 2,"reqId":'
                            ' 1597766666168851000,"signatures": {"LjgpST2rjsox":'
                            ' "4ATKMn6Y9sTgwqaGTm7py2c2M8x1EVDTWKZArwyuPgjU"},'
                            ' "taaAcceptance": {"mechanism": "manual","taaDigest":'
                            ' "f50fe2c2ab977006761d36bd6f23e4c6a7e0fc2feb9f62","time":'
                            " 1597708800,}}"
                        )
                    },
                }
            }
        ),
        required=False,
    )
    meta_data = fields.Dict(
        required=False,
        metadata={
            "example": {
                "context": {"param1": "param1_value", "param2": "param2_value"},
                "post_process": [{"topic": "topic_value", "other": "other_value"}],
            }
        },
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread Identifier", "example": UUID4_EXAMPLE},
    )
    connection_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "The connection identifier for this particular transaction record"
            ),
            "example": UUID4_EXAMPLE,
        },
    )
    endorser_write_txn = fields.Boolean(
        required=False,
        metadata={
            "description": (
                "Request Endorser to write the ledger transaction, "
                "this parameter is deprecated and no longer supported."
            ),
            "example": False,
        },
    )



"""The timing decorator (~timing).

This decorator allows the timing of agent messages to be communicated
and constrained.
"""

from datetime import datetime
from typing import Optional, Union

from marshmallow import EXCLUDE, fields

from ..models.base import BaseModel, BaseModelSchema
from ..util import datetime_to_str
from ..valid import ISO8601_DATETIME_EXAMPLE, ISO8601_DATETIME_VALIDATE


class TimingDecorator(BaseModel):
    """Class representing the timing decorator."""

    class Meta:
        """TimingDecorator metadata."""

        schema_class = "TimingDecoratorSchema"

    def __init__(
        self,
        *,
        in_time: Union[str, datetime] = None,
        out_time: Union[str, datetime] = None,
        stale_time: Union[str, datetime] = None,
        expires_time: Union[str, datetime] = None,
        delay_milli: Optional[int] = None,
        wait_until_time: Union[str, datetime] = None,
    ):
        """Initialize a TimingDecorator instance.

        Args:
            in_time: The time the message was received
            out_time: The time the message was dispatched
            stale_time: When the message should be considered stale
            expires_time: When the message should be considered expired
            delay_milli: The number of milliseconds to delay processing
            wait_until_time: The earliest time at which to perform processing
        """
        super().__init__()
        self.in_time = datetime_to_str(in_time)
        self.out_time = datetime_to_str(out_time)
        self.stale_time = datetime_to_str(stale_time)
        self.expires_time = datetime_to_str(expires_time)
        self.delay_milli = delay_milli
        self.wait_until_time = datetime_to_str(wait_until_time)


class TimingDecoratorSchema(BaseModelSchema):
    """Timing decorator schema used in serialization/deserialization."""

    class Meta:
        """TimingDecoratorSchema metadata."""

        model_class = TimingDecorator
        unknown = EXCLUDE

    in_time = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Time of message receipt",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )
    out_time = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Time of message dispatch",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )
    stale_time = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Time when message should be considered stale",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )
    expires_time = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Time when message should be considered expired",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )
    delay_milli = fields.Int(
        required=False,
        metadata={
            "description": "Number of milliseconds to delay processing",
            "example": 1000,
            "strict": True,
        },
    )
    wait_until_time = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Earliest time at which to perform processing",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )



import os
import shutil
from unittest import IsolatedAsyncioTestCase

import pytest
from aiohttp.web import HTTPNotFound

from ...admin.request_context import AdminRequestContext
from ...anoncreds.models.revocation import RevRegDef, RevRegDefValue
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module


class TestRevocationRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
                "wallet.type": "askar-anoncreds",
            },
        )
        self.context = self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

        self.test_did = "sample-did"

    async def test_validate_cred_rev_rec_qs_and_revoke_req(self):
        for req in (
            test_module.CredRevRecordQueryStringSchema(),
            test_module.RevokeRequestSchemaAnonCreds(),
        ):
            req.validate_fields(
                {
                    "rev_reg_id": test_module.ANONCREDS_REV_REG_ID_EXAMPLE,
                    "cred_rev_id": test_module.ANONCREDS_CRED_REV_ID_EXAMPLE,
                }
            )
            req.validate_fields({"cred_ex_id": test_module.UUID4_EXAMPLE})
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields({})
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {"rev_reg_id": test_module.ANONCREDS_REV_REG_ID_EXAMPLE}
                )
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {"cred_rev_id": test_module.ANONCREDS_CRED_REV_ID_EXAMPLE}
                )
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {
                        "rev_reg_id": test_module.ANONCREDS_REV_REG_ID_EXAMPLE,
                        "cred_ex_id": test_module.UUID4_EXAMPLE,
                    }
                )
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {
                        "cred_rev_id": test_module.ANONCREDS_CRED_REV_ID_EXAMPLE,
                        "cred_ex_id": test_module.UUID4_EXAMPLE,
                    }
                )
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {
                        "rev_reg_id": test_module.ANONCREDS_REV_REG_ID_EXAMPLE,
                        "cred_rev_id": test_module.ANONCREDS_CRED_REV_ID_EXAMPLE,
                        "cred_ex_id": test_module.UUID4_EXAMPLE,
                    }
                )

    async def test_revoke(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock()

            await test_module.revoke(self.request)

            mock_response.assert_called_once_with({})

    async def test_revoke_by_cred_ex_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "cred_ex_id": "dummy-cxid",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock()

            await test_module.revoke(self.request)

            mock_response.assert_called_once_with({})

    async def test_revoke_not_found(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.revoke(self.request)

    async def test_publish_revocations(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            pub_pending = mock.CoroutineMock()
            mock_mgr.return_value.publish_pending_revocations = pub_pending

            await test_module.publish_revocations(self.request)

            mock_response.assert_called_once_with({"rrid2crid": pub_pending.return_value})

    async def test_publish_revocations_x(self):
        self.request.json = mock.CoroutineMock()

        with mock.patch.object(
            test_module, "RevocationManager", autospec=True
        ) as mock_mgr:
            pub_pending = mock.CoroutineMock(side_effect=test_module.RevocationError())
            mock_mgr.return_value.publish_pending_revocations = pub_pending

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.publish_revocations(self.request)

    async def test_rev_regs_created(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.query = {
            "cred_def_id": CRED_DEF_ID,
            "state": test_module.IssuerRevRegRecord.STATE_ACTIVE,
        }

        with (
            mock.patch.object(
                test_module.AnonCredsRevocation,
                "get_created_revocation_registry_definitions",
                mock.AsyncMock(),
            ) as mock_query,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_query.return_value = ["dummy"]

            result = await test_module.get_rev_regs(self.request)
            mock_json_response.assert_called_once_with({"rev_reg_ids": ["dummy"]})
            assert result is mock_json_response.return_value

    async def test_get_rev_reg(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        RECORD_ID = "4ba81d6e-f341-4e37-83d4-6b1d3e25a7bd"
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "AnonCredsRevocation", autospec=True
            ) as mock_anon_creds_revoc,
            mock.patch.object(test_module, "uuid4", mock.Mock()) as mock_uuid,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_uuid.return_value = RECORD_ID
            mock_anon_creds_revoc.return_value = mock.MagicMock(
                get_created_revocation_registry_definition=mock.AsyncMock(
                    return_value=RevRegDef(
                        issuer_id="issuer_id",
                        type="CL_ACCUM",
                        cred_def_id="cred_def_id",
                        tag="tag",
                        value=RevRegDefValue(
                            public_keys={},
                            max_cred_num=100,
                            tails_hash="tails_hash",
                            tails_location="tails_location",
                        ),
                    )
                ),
                get_created_revocation_registry_definition_state=mock.AsyncMock(
                    return_value=test_module.RevRegDefState.STATE_FINISHED
                ),
                get_pending_revocations=mock.AsyncMock(return_value=[]),
            )

            result = await test_module.get_rev_reg(self.request)
            mock_json_response.assert_called_once_with(
                {
                    "result": {
                        "tails_local_path": "tails_location",
                        "tails_hash": "tails_hash",
                        "state": test_module.RevRegDefState.STATE_FINISHED,
                        "issuer_did": "issuer_id",
                        "pending_pub": [],
                        "revoc_reg_def": {
                            "ver": "1.0",
                            "id": REV_REG_ID,
                            "revocDefType": "CL_ACCUM",
                            "tag": "tag",
                            "credDefId": "cred_def_id",
                        },
                        "max_cred_num": 100,
                        "record_id": RECORD_ID,
                        "tag": "tag",
                        "revoc_def_type": "CL_ACCUM",
                        "revoc_reg_id": REV_REG_ID,
                        "cred_def_id": "cred_def_id",
                    }
                }
            )
            assert result is mock_json_response.return_value

    async def test_get_rev_reg_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "AnonCredsRevocation", autospec=True
            ) as mock_anon_creds_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_anon_creds_revoc.return_value = mock.MagicMock(
                get_created_revocation_registry_definition=mock.AsyncMock(
                    return_value=None
                ),
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.get_rev_reg(self.request)
            mock_json_response.assert_not_called()

    async def test_get_rev_reg_issued(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module.AnonCredsRevocation,
                "get_created_revocation_registry_definition",
                autospec=True,
            ) as mock_rev_reg_def,
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "query_by_ids",
                mock.CoroutineMock(),
            ) as mock_query,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_rev_reg_def.return_value = {}
            mock_query.return_value = [{}, {}]
            result = await test_module.get_rev_reg_issued_count(self.request)

            mock_json_response.assert_called_once_with({"result": 2})
            assert result is mock_json_response.return_value

    async def test_get_rev_reg_issued_x(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with mock.patch.object(
            test_module.AnonCredsRevocation,
            "get_created_revocation_registry_definition",
            autospec=True,
        ) as mock_rev_reg_def:
            mock_rev_reg_def.return_value = None

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.get_rev_reg_issued(self.request)

    async def test_get_cred_rev_record(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        CRED_REV_ID = "1"

        self.request.query = {
            "rev_reg_id": REV_REG_ID,
            "cred_rev_id": CRED_REV_ID,
        }

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_ids",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value="dummy")
            )
            result = await test_module.get_cred_rev_record(self.request)

            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_get_cred_rev_record_by_cred_ex_id(self):
        CRED_EX_ID = test_module.UUID4_EXAMPLE

        self.request.query = {"cred_ex_id": CRED_EX_ID}

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_cred_ex_id",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value="dummy")
            )
            result = await test_module.get_cred_rev_record(self.request)

            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_get_cred_rev_record_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        CRED_REV_ID = "1"

        self.request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": REV_REG_ID,
                "cred_rev_id": CRED_REV_ID,
            }
        )

        with mock.patch.object(
            test_module.IssuerCredRevRecord,
            "retrieve_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            mock_retrieve.side_effect = test_module.StorageNotFoundError("no such rec")
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.get_cred_rev_record(self.request)

    async def test_get_tails_file(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module.AnonCredsRevocation,
                "get_created_revocation_registry_definition",
                mock.AsyncMock(),
            ) as mock_get_rev_reg,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_file_response,
        ):
            mock_get_rev_reg.return_value = RevRegDef(
                issuer_id="issuer_id",
                type="CL_ACCUM",
                cred_def_id="cred_def_id",
                tag="tag",
                value=RevRegDefValue(
                    public_keys={},
                    max_cred_num=100,
                    tails_hash="tails_hash",
                    tails_location="tails_location",
                ),
            )

            result = await test_module.get_tails_file(self.request)
            mock_file_response.assert_called_once_with(path="tails_location", status=200)
            assert result is mock_file_response.return_value

    async def test_get_tails_file_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module.AnonCredsRevocation,
                "get_created_revocation_registry_definition",
                mock.AsyncMock(),
            ) as mock_get_rev_reg,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_file_response,
        ):
            mock_get_rev_reg.return_value = None

            with self.assertRaises(HTTPNotFound):
                await test_module.get_tails_file(self.request)
            mock_file_response.assert_not_called()

    async def test_set_rev_reg_state(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        RECORD_ID = "4ba81d6e-f341-4e37-83d4-6b1d3e25a7bd"
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        self.request.query = {
            "state": test_module.RevRegDefState.STATE_FINISHED,
        }

        with (
            mock.patch.object(
                test_module, "AnonCredsRevocation", autospec=True
            ) as mock_anon_creds_revoc,
            mock.patch.object(test_module, "uuid4", mock.Mock()) as mock_uuid,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_uuid.return_value = RECORD_ID
            mock_anon_creds_revoc.return_value = mock.MagicMock(
                set_rev_reg_state=mock.AsyncMock(return_value={}),
                get_created_revocation_registry_definition=mock.AsyncMock(
                    return_value=RevRegDef(
                        issuer_id="issuer_id",
                        type="CL_ACCUM",
                        cred_def_id="cred_def_id",
                        tag="tag",
                        value=RevRegDefValue(
                            public_keys={},
                            max_cred_num=100,
                            tails_hash="tails_hash",
                            tails_location="tails_location",
                        ),
                    )
                ),
                get_created_revocation_registry_definition_state=mock.AsyncMock(
                    return_value=test_module.RevRegDefState.STATE_FINISHED
                ),
                get_pending_revocations=mock.AsyncMock(return_value=[]),
            )

            result = await test_module.set_rev_reg_state(self.request)
            mock_json_response.assert_called_once_with(
                {
                    "result": {
                        "tails_local_path": "tails_location",
                        "tails_hash": "tails_hash",
                        "state": test_module.RevRegDefState.STATE_FINISHED,
                        "issuer_did": "issuer_id",
                        "pending_pub": [],
                        "revoc_reg_def": {
                            "ver": "1.0",
                            "id": REV_REG_ID,
                            "revocDefType": "CL_ACCUM",
                            "tag": "tag",
                            "credDefId": "cred_def_id",
                        },
                        "max_cred_num": 100,
                        "record_id": RECORD_ID,
                        "tag": "tag",
                        "revoc_def_type": "CL_ACCUM",
                        "revoc_reg_id": REV_REG_ID,
                        "cred_def_id": "cred_def_id",
                    }
                }
            )
            assert result is mock_json_response.return_value

    async def test_set_rev_reg_state_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        self.request.query = {
            "state": test_module.RevRegDefState.STATE_FINISHED,
        }

        with (
            mock.patch.object(
                test_module.AnonCredsRevocation,
                "get_created_revocation_registry_definition",
                mock.AsyncMock(),
            ) as mock_rev_reg_def,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_rev_reg_def.return_value = None

            with self.assertRaises(HTTPNotFound):
                await test_module.set_rev_reg_state(self.request)
            mock_json_response.assert_not_called()

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(
            _state={
                "swagger_dict": {
                    "paths": {
                        "/revocation/registry/{rev_reg_id}/tails-file": {
                            "get": {"responses": {"200": {"description": "tails file"}}}
                        }
                    }
                }
            }
        )
        test_module.post_process_routes(mock_app)
        assert mock_app._state["swagger_dict"]["paths"][
            "/revocation/registry/{rev_reg_id}/tails-file"
        ]["get"]["responses"]["200"]["schema"] == {"type": "string", "format": "binary"}

        assert "tags" in mock_app._state["swagger_dict"]


class TestDeleteTails(IsolatedAsyncioTestCase):
    def setUp(self):
        self.rev_reg_id = "rev_reg_id_123"
        self.cred_def_id = "cred_def_id_456"

        self.main_dir_rev = "path/to/main/dir/rev"
        self.tails_path = os.path.join(self.main_dir_rev, "tails")
        if not (os.path.exists(self.main_dir_rev)):
            os.makedirs(self.main_dir_rev)
        open(self.tails_path, "w").close()

    @pytest.mark.xfail(reason="This test never worked but was skipped due to a bug")
    async def test_delete_tails_by_rev_reg_id(self):
        # Setup
        rev_reg_id = self.rev_reg_id

        # Test
        result = await test_module.delete_tails(
            {"context": None, "query": {"rev_reg_id": rev_reg_id}}
        )

        # Assert
        self.assertEqual(result, {"message": "All files deleted successfully"})
        self.assertFalse(os.path.exists(self.tails_path))

    @pytest.mark.xfail(reason="This test never worked but was skipped due to a bug")
    async def test_delete_tails_by_cred_def_id(self):
        # Setup
        cred_def_id = self.cred_def_id
        main_dir_cred = "path/to/main/dir/cred"
        os.makedirs(main_dir_cred)
        cred_dir = os.path.join(main_dir_cred, cred_def_id)
        os.makedirs(cred_dir)

        # Test
        result = await test_module.delete_tails(
            {"context": None, "query": {"cred_def_id": cred_def_id}}
        )

        # Assert
        self.assertEqual(result, {"message": "All files deleted successfully"})
        self.assertFalse(os.path.exists(cred_dir))
        self.assertTrue(os.path.exists(main_dir_cred))

    @pytest.mark.xfail(reason="This test never worked but was skipped due to a bug")
    async def test_delete_tails_not_found(self):
        # Setup
        cred_def_id = "invalid_cred_def_id"

        # Test
        result = await test_module.delete_tails(
            {"context": None, "query": {"cred_def_id": cred_def_id}}
        )

        # Assert
        self.assertEqual(result, {"message": "No such file or directory"})
        self.assertTrue(os.path.exists(self.main_dir_rev))

    def tearDown(self):
        if os.path.exists(self.main_dir_rev):
            shutil.rmtree(self.main_dir_rev)



"""Multicodec wrap and unwrap functions."""

from enum import Enum
from typing import Literal, NamedTuple, Optional, Union


class Multicodec(NamedTuple):
    """Multicodec base class."""

    name: str
    code: bytes


class SupportedCodecs(Enum):
    """Enumeration of supported multicodecs."""

    ed25519_pub = Multicodec("ed25519-pub", b"\xed\x01")
    x25519_pub = Multicodec("x25519-pub", b"\xec\x01")
    bls12381g1 = Multicodec("bls12_381-g1-pub", b"\xea\x01")
    bls12381g2 = Multicodec("bls12_381-g2-pub", b"\xeb\x01")
    bls12381g1g2 = Multicodec("bls12_381-g1g2-pub", b"\xee\x01")
    secp256k1_pub = Multicodec("secp256k1-pub", b"\xe7\x01")
    p256_pub = Multicodec("p256-pub", b"\x80\x24")

    @classmethod
    def by_name(cls, name: str) -> Multicodec:
        """Get multicodec by name."""
        for codec in cls:
            if codec.value.name == name:
                return codec.value
        raise ValueError(f"Unsupported multicodec: {name}")

    @classmethod
    def for_data(cls, data: bytes) -> Multicodec:
        """Get multicodec by data."""
        for codec in cls:
            if data.startswith(codec.value.code):
                return codec.value
        raise ValueError("Unsupported multicodec")


MulticodecStr = Literal[
    "ed25519-pub",
    "x25519-pub",
    "bls12_381-g1-pub",
    "bls12_381-g2-pub",
    "bls12_381-g1g2-pub",
    "secp256k1-pub",
    "p256-pub",
]


def multicodec(name: str) -> Multicodec:
    """Get multicodec by name."""
    return SupportedCodecs.by_name(name)


def wrap(multicodec: Union[Multicodec, MulticodecStr], data: bytes) -> bytes:
    """Wrap data with multicodec prefix."""
    if isinstance(multicodec, str):
        multicodec = SupportedCodecs.by_name(multicodec)
    elif isinstance(multicodec, Multicodec):
        pass
    else:
        raise TypeError("multicodec must be Multicodec or MulticodecStr")

    return multicodec.code + data


def unwrap(data: bytes, codec: Optional[Multicodec] = None) -> tuple[Multicodec, bytes]:
    """Unwrap data with multicodec prefix."""
    if not codec:
        codec = SupportedCodecs.for_data(data)
    return codec, data[len(codec.code) :]



"""Common exception classes."""

import re
from typing import Optional


class BaseError(Exception):
    """Generic exception class which other exceptions should inherit from."""

    def __init__(self, *args, error_code: Optional[str] = None, **kwargs):
        """Initialize a BaseError instance."""
        super().__init__(*args, **kwargs)
        self.error_code = error_code if error_code else None

    @property
    def message(self) -> str:
        """Accessor for the error message."""
        return str(self.args[0]).strip() if self.args else ""

    @property
    def roll_up(self) -> str:
        """Accessor for nested error messages rolled into one line.

        For display: aiohttp.web errors truncate after newline.
        """

        def flatten(exc: Exception):
            return ".".join(
                (
                    re.sub(
                        r"\n\s*",
                        ". ",
                        (
                            str(exc.args[0]).strip()
                            if exc.args
                            else exc.__class__.__name__
                        ),
                    ).strip()
                ).rsplit(".", 1)
            )

        line = flatten(self)
        err = self
        while err.__cause__:
            err = err.__cause__
            line += ". {}".format(flatten(err))
        return f"{line.strip()}."


class ProfileError(BaseError):
    """Base error for profile operations."""


class ProfileDuplicateError(ProfileError):
    """Profile with the given name already exists."""


class ProfileNotFoundError(ProfileError):
    """Requested profile was not found."""


class ProfileSessionInactiveError(ProfileError):
    """Error raised when a profile session is not currently active."""


class StartupError(BaseError):
    """Error raised when there is a problem starting the conductor."""


class ProtocolDefinitionValidationError(BaseError):
    """Error raised when there is a problem validating a protocol definition."""


class ProtocolMinorVersionNotSupported(BaseError):
    """Minimum minor version protocol error.

    Error raised when protocol support exists
    but minimum minor version is higher than in @type parameter.
    """



from unittest import TestCase, mock

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import DISCLOSURES, PROTOCOL_PACKAGE
from ..disclosures import Disclosures


class TestDisclosures(TestCase):
    test_disclosure_a = [
        {
            "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
            "feature-type": "protocol",
            "roles": [],
        }
    ]
    test_disclosure_b = [
        {
            "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
            "feature-type": "protocol",
            "roles": [],
        },
        {"feature-type": "goal-code", "id": "aries.sell.goods.consumer"},
    ]

    def test_init(self):
        disclosures = Disclosures(disclosures=self.test_disclosure_a)
        assert disclosures.disclosures == self.test_disclosure_a

        disclosures = Disclosures(disclosures=self.test_disclosure_b)
        assert disclosures.disclosures == self.test_disclosure_b

    def test_type(self):
        disclosures = Disclosures(disclosures=self.test_disclosure_a)
        assert disclosures._type == DIDCommPrefix.qualify_current(DISCLOSURES)

        disclosures = Disclosures(disclosures=self.test_disclosure_b)
        assert disclosures._type == DIDCommPrefix.qualify_current(DISCLOSURES)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.disclosures.DisclosuresSchema.load")
    def test_deserialize(self, mock_disclosures_schema_load):
        obj = {"obj": "obj"}

        disclosures = Disclosures.deserialize(obj)
        mock_disclosures_schema_load.assert_called_once_with(obj)

        assert disclosures is mock_disclosures_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.disclosures.DisclosuresSchema.dump")
    def test_serialize(self, mock_disclosures_schema_dump):
        disclosures = Disclosures(disclosures=self.test_disclosure_b)

        disclosures_dict = disclosures.serialize()
        mock_disclosures_schema_dump.assert_called_once_with(disclosures)

        assert disclosures_dict is mock_disclosures_schema_dump.return_value


class TestDiscloseSchema(TestCase):
    test_disclosures = [
        {
            "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
            "feature-type": "protocol",
            "roles": [],
        },
        {"feature-type": "goal-code", "id": "aries.sell.goods.consumer"},
    ]
    disclosures = Disclosures(disclosures=test_disclosures)
    test_invalid_disclosures = [
        {
            "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
            "feature-type": "protocol",
            "roles": [],
        },
        {"test": "test"},
    ]
    invalid_disclosures = Disclosures(disclosures=test_invalid_disclosures)

    def test_make_model(self):
        data = self.disclosures.serialize()
        model_instance = Disclosures.deserialize(data)
        assert isinstance(model_instance, Disclosures)

        with self.assertRaises(BaseModelError):
            data = self.invalid_disclosures.serialize()
            model_instance = Disclosures.deserialize(data)



"""AnonCreds cred def OpenAPI validators."""

from typing import Optional

from anoncreds import CredentialDefinition
from marshmallow import EXCLUDE, fields
from marshmallow.validate import OneOf
from typing_extensions import Literal

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_DID_EXAMPLE,
    ANONCREDS_SCHEMA_ID_EXAMPLE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
)

NUM_STR_WHOLE = {
    "validate": NUM_STR_WHOLE_VALIDATE,
    "metadata": {"example": NUM_STR_WHOLE_EXAMPLE},
}


class CredDefValuePrimary(BaseModel):
    """PrimarySchema."""

    class Meta:
        """PrimarySchema metadata."""

        schema_class = "CredDefValuePrimarySchemaAnonCreds"

    def __init__(self, n: str, s: str, r: dict, rctxt: str, z: str, **kwargs):
        """Initialize an instance.

        Args:
            n: is a safe RSA-2048 number.
            s: is a randomly selected quadratic residue of n.
            r: is an object that defines a CL-RSA public key fragment
                for each attribute in the credential.
            rctxt: is equal to s^(xrctxt), where xrctxt is a randomly selected integer
                between 2 and p'q'-1.
            z: is equal to s^(xz), where xz is a randomly selected integer between 2
                and p'q'-1. This makes up part of the CL-RSA public key, independent of
                the message blocks being signed.
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.n = n
        self.s = s
        self.r = r
        self.rctxt = rctxt
        self.z = z


class CredDefValuePrimarySchemaAnonCreds(BaseModelSchema):
    """Cred def value primary schema."""

    class Meta:
        """CredDefValuePrimarySchema metadata."""

        model_class = CredDefValuePrimary
        unknown = EXCLUDE

    n = fields.Str(**NUM_STR_WHOLE)
    s = fields.Str(**NUM_STR_WHOLE)
    r = fields.Dict()
    rctxt = fields.Str(**NUM_STR_WHOLE)
    z = fields.Str(**NUM_STR_WHOLE)


class CredDefValueRevocation(BaseModel):
    """Cred def value revocation."""

    class Meta:
        """CredDefValueRevocation metadata."""

        schema_class = "CredDefValueRevocationSchemaAnonCreds"

    def __init__(
        self,
        g: str,
        g_dash: str,
        h: str,
        h0: str,
        h1: str,
        h2: str,
        htilde: str,
        h_cap: str,
        u: str,
        pk: str,
        y: str,
    ):
        """Initialize an instance.

        Args:
            g: is a generator for the elliptic curve group G1.
            g_dash: is a generator for the elliptic curve group G2.
            h: is an elliptic curve point selected uniformly at random from G1.
            h0: is an elliptic curve point selected uniformly at random from G1.
            h1: is an elliptic curve point selected uniformly at random from G1.
            h2: is an elliptic curve point selected uniformly at random from G1.
            htilde: is an elliptic curve point selected uniformly at random from G1.
            h_cap: is an elliptic curve point selected uniformly at random from G2.
            u: is an elliptic curve point selected uniformly at random from G2.
            pk: is the public key in G1 for the issuer with respect to this accumulator,
                computed as g^sk (in multiplicative notation), where sk is from
                r_key above.
            y: is the an elliptic curve point in G2. computed as h_cap^x
                (in multiplicative notation), where x is from r_key above
        """
        self.g = g
        self.g_dash = g_dash
        self.h = h
        self.h0 = h0
        self.h1 = h1
        self.h2 = h2
        self.htilde = htilde
        self.h_cap = h_cap
        self.u = u
        self.pk = pk
        self.y = y


class CredDefValueRevocationSchemaAnonCreds(BaseModelSchema):
    """Cred def value revocation schema."""

    class Meta:
        """Metadata."""

        model_class = CredDefValueRevocation
        unknown = EXCLUDE

    g = fields.Str(metadata={"example": "1 1F14F&ECB578F 2 095E45DDF417D"})
    g_dash = fields.Str(
        metadata={"example": "1 1D64716fCDC00C 1 0C781960FA66E3D3 2 095E45DDF417D"}
    )
    h = fields.Str(metadata={"example": "1 16675DAE54BFAE8 2 095E45DD417D"})
    h0 = fields.Str(metadata={"example": "1 21E5EF9476EAF18 2 095E45DDF417D"})
    h1 = fields.Str(metadata={"example": "1 236D1D99236090 2 095E45DDF417D"})
    h2 = fields.Str(metadata={"example": "1 1C3AE8D1F1E277 2 095E45DDF417D"})
    htilde = fields.Str(metadata={"example": "1 1D8549E8C0F8 2 095E45DDF417D"})
    h_cap = fields.Str(
        metadata={"example": "1 1B2A32CF3167 1 2490FEBF6EE55 1 0000000000000000"}
    )
    u = fields.Str(
        metadata={"example": "1 0C430AAB2B4710 1 1CB3A0932EE7E 1 0000000000000000"}
    )
    pk = fields.Str(
        metadata={"example": "1 142CD5E5A7DC 1 153885BD903312 2 095E45DDF417D"}
    )
    y = fields.Str(
        metadata={"example": "1 153558BD903312 2 095E45DDF417D 1 0000000000000000"}
    )


class CredDefValue(BaseModel):
    """Cred def value."""

    class Meta:
        """CredDefValue metadata."""

        schema_class = "CredDefValueSchemaAnonCreds"

    def __init__(
        self,
        primary: CredDefValuePrimary,
        revocation: Optional[CredDefValueRevocation] = None,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            primary: Cred Def value primary
            revocation: Cred Def value revocation
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.primary = primary
        self.revocation = revocation


class CredDefValueSchemaAnonCreds(BaseModelSchema):
    """Cred def value schema."""

    class Meta:
        """CredDefValueSchema metadata."""

        model_class = CredDefValue
        unknown = EXCLUDE

    primary = fields.Nested(
        CredDefValuePrimarySchemaAnonCreds(),
        metadata={"description": "Primary value for credential definition"},
    )
    revocation = fields.Nested(
        CredDefValueRevocationSchemaAnonCreds(),
        metadata={"description": "Revocation value for credential definition"},
        required=False,
    )


class CredDef(BaseModel):
    """AnonCredsCredDef."""

    class Meta:
        """AnonCredsCredDef metadata."""

        schema_class = "CredDefSchema"

    def __init__(
        self,
        issuer_id: str,
        schema_id: str,
        type: Literal["CL"],
        tag: str,
        value: CredDefValue,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            issuer_id: Issuer ID
            schema_id: Schema ID
            type: Type
            tag: Tag
            value: Cred Def value
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.issuer_id = issuer_id
        self.schema_id = schema_id
        self.type = type
        self.tag = tag
        self.value = value

    @classmethod
    def from_native(cls, cred_def: CredentialDefinition):
        """Convert a native credential definition to a CredDef object."""
        return cls.deserialize(cred_def.to_json())

    def to_native(self) -> CredentialDefinition:
        """Convert to native anoncreds credential definition."""
        return CredentialDefinition.load(self.serialize())


class CredDefSchema(BaseModelSchema):
    """CredDefSchema."""

    class Meta:
        """CredDefSchema metadata."""

        model_class = CredDef
        unknown = EXCLUDE

    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition or schema",
            "example": ANONCREDS_DID_EXAMPLE,
        },
        data_key="issuerId",
    )
    schema_id = fields.Str(
        data_key="schemaId",
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        },
    )
    type = fields.Str(validate=OneOf(["CL"]))
    tag = fields.Str(
        metadata={
            "description": (
                "The tag value passed in by the Issuer to "
                "an AnonCred's Credential Definition create and store implementation."
            ),
            "example": "default",
        }
    )
    value = fields.Nested(CredDefValueSchemaAnonCreds())


class CredDefState(BaseModel):
    """CredDefState."""

    STATE_FINISHED = "finished"
    STATE_FAILED = "failed"
    STATE_ACTION = "action"
    STATE_WAIT = "wait"

    class Meta:
        """CredDefState metadata."""

        schema_class = "CredDefStateSchema"

    def __init__(
        self,
        state: str,
        credential_definition_id: Optional[str],
        credential_definition: CredDef,
    ):
        """Initialize an instance.

        Args:
            state: State
            credential_definition_id: Cred Def ID
            credential_definition: Cred Def
        """
        self.state = state
        self.credential_definition_id = credential_definition_id
        self.credential_definition = credential_definition


class CredDefStateSchema(BaseModelSchema):
    """CredDefStateSchema."""

    class Meta:
        """CredDefStateSchema metadata."""

        model_class = CredDefState
        unknown = EXCLUDE

    state = fields.Str(
        validate=OneOf(
            [
                CredDefState.STATE_FINISHED,
                CredDefState.STATE_FAILED,
                CredDefState.STATE_ACTION,
                CredDefState.STATE_WAIT,
            ]
        )
    )
    credential_definition_id = fields.Str(
        metadata={
            "description": "credential definition id",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
        allow_none=True,
    )
    credential_definition = fields.Nested(
        CredDefSchema(), metadata={"description": "credential definition"}
    )


class CredDefResult(BaseModel):
    """Cred def result."""

    class Meta:
        """CredDefResult metadata."""

        schema_class = "CredDefResultSchema"

    def __init__(
        self,
        job_id: Optional[str],
        credential_definition_state: CredDefState,
        registration_metadata: dict,
        credential_definition_metadata: dict,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            job_id: Job ID
            credential_definition_state: Cred Def state
            registration_metadata: Registration metadata
            credential_definition_metadata: Cred Def metadata
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.job_id = job_id
        self.credential_definition_state = credential_definition_state
        self.registration_metadata = registration_metadata
        self.credential_definition_metadata = credential_definition_metadata


class CredDefResultSchema(BaseModelSchema):
    """Cred def result schema."""

    class Meta:
        """CredDefResultSchema metadata."""

        model_class = CredDefResult
        unknown = EXCLUDE

    job_id = fields.Str()
    credential_definition_state = fields.Nested(CredDefStateSchema())
    registration_metadata = fields.Dict()
    # For indy, credential_definition_metadata will contain the seqNo
    credential_definition_metadata = fields.Dict()


class GetCredDefResult(BaseModel):
    """Get cred def result."""

    class Meta:
        """AnonCredsRegistryGetCredDef metadata."""

        schema_class = "GetCredDefResultSchema"

    def __init__(
        self,
        credential_definition_id: str,
        credential_definition: CredDef,
        resolution_metadata: dict,
        credential_definition_metadata: dict,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            credential_definition_id: Cred Def ID
            credential_definition: Cred Def
            resolution_metadata: Resolution metadata
            credential_definition_metadata: Cred Def metadata
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.credential_definition_id = credential_definition_id
        self.credential_definition = credential_definition
        self.resolution_metadata = resolution_metadata
        self.credential_definition_metadata = credential_definition_metadata


class GetCredDefResultSchema(BaseModelSchema):
    """GetCredDefResultSchema."""

    class Meta:
        """GetCredDefResultSchema metadata."""

        model_class = GetCredDefResult
        unknown = EXCLUDE

    credential_definition_id = fields.Str(
        metadata={
            "description": "credential definition id",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )
    credential_definition = fields.Nested(
        CredDefSchema(), metadata={"description": "credential definition"}
    )
    resolution_metadata = fields.Dict()
    credential_definitions_metadata = fields.Dict()



"""Message type identifiers for Revocation Notification protocol."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/blob/main/features/"
    "0721-revocation-notification-v2/README.md"
)
PROTOCOL = "revocation_notification"
VERSION = "2.0"
BASE = f"{PROTOCOL}/{VERSION}"

# Message types
REVOKE = f"{BASE}/revoke"

PROTOCOL_PACKAGE = "acapy_agent.protocols.revocation_notification.v2_0"
MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {REVOKE: f"{PROTOCOL_PACKAGE}.messages.revoke.Revoke"}
)



"""Request context class.

A request context provides everything required by handlers and other parts
of the system to process a message.
"""

from typing import Mapping, Optional, Type

from ..config.injection_context import InjectionContext
from ..config.injector import Injector, InjectType
from ..config.settings import Settings
from ..connections.models.conn_record import ConnRecord
from ..core.profile import Profile, ProfileSession
from ..transport.inbound.receipt import MessageReceipt
from .agent_message import AgentMessage


class RequestContext:
    """Context established by the Conductor and passed into message handlers."""

    def __init__(
        self,
        profile: Profile,
        *,
        context: Optional[InjectionContext] = None,
        settings: Optional[Mapping[str, object]] = None,
    ):
        """Initialize an instance of RequestContext."""
        self._connection_ready = False
        self._connection_record = None
        self._context = (context or profile.context).start_scope(settings)
        self._message = None
        self._message_receipt = None
        self._profile = profile

    @property
    def connection_ready(self) -> bool:
        """Accessor for the flag indicating an active connection with the sender.

        Returns:
            True if the connection is active, else False

        """
        return self._connection_ready

    @connection_ready.setter
    def connection_ready(self, active: bool):
        """Setter for the flag indicating an active connection with the sender.

        Args:
            active: The new active value

        """
        self._connection_ready = active

    @property
    def connection_record(self) -> Optional[ConnRecord]:
        """Accessor for the related connection record."""
        return self._connection_record

    @connection_record.setter
    def connection_record(self, record: ConnRecord):
        """Setter for the related connection record.

        :param record: ConnRecord:

        """
        self._connection_record = record

    @property
    def default_endpoint(self) -> str:
        """Accessor for the default agent endpoint (from agent config).

        Returns:
            The default agent endpoint

        """
        return self._context.settings.get("default_endpoint")

    @default_endpoint.setter
    def default_endpoint(self, endpoint: str):
        """Setter for the default agent endpoint (from agent config).

        Args:
            endpoint: The new default endpoint

        """
        self._context.settings["default_endpoint"] = endpoint

    @property
    def default_label(self) -> str:
        """Accessor for the default agent label (from agent config).

        Returns:
            The default label

        """
        return self._context.settings["default_label"]

    @default_label.setter
    def default_label(self, label: str):
        """Setter for the default agent label (from agent config).

        Args:
            label: The new default label

        """
        self._context.settings["default_label"] = label

    @property
    def message(self) -> AgentMessage:
        """Accessor for the deserialized message instance.

        Returns:
            This context's agent message

        """
        return self._message

    @message.setter
    def message(self, msg: AgentMessage):
        """Setter for the deserialized message instance.

        Args:
            msg: This context's new agent message
        """
        self._message = msg

    @property
    def message_receipt(self) -> MessageReceipt:
        """Accessor for the message receipt information.

        Returns:
            This context's message receipt information

        """
        return self._message_receipt

    @message_receipt.setter
    def message_receipt(self, receipt: MessageReceipt):
        """Setter for the message receipt information.

        Args:
            receipt: This context's new message receipt information
        """
        self._message_receipt = receipt

    @property
    def injector(self) -> Injector:
        """Accessor for the associated `Injector` instance."""
        return self._context.injector

    @property
    def profile(self) -> Profile:
        """Accessor for the associated `Profile` instance."""
        return self._profile

    @property
    def settings(self) -> Settings:
        """Accessor for the context settings."""
        return self._context.settings

    def session(self) -> ProfileSession:
        """Start a new interactive session with no transaction support requested."""
        return self.profile.session(self._context)

    def transaction(self) -> ProfileSession:
        """Start a new interactive session with commit and rollback support.

        If the current backend does not support transactions, then commit
        and rollback operations of the session will not have any effect.
        """
        return self.profile.transaction(self._context)

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional mapping providing configuration to the provider

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject(base_cls, settings)

    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject_or(base_cls, settings, default)

    def update_settings(self, settings: Mapping[str, object]):
        """Update the scope with additional settings."""
        self._context.update_settings(settings)

    @classmethod
    def test_context(cls, profile) -> "RequestContext":
        """Quickly set up a new request context for tests."""
        return RequestContext(profile)

    def __repr__(self) -> str:
        """Provide a human readable representation of this object.

        Returns:
            A human readable representation of this object

        """
        skip = ()
        items = (
            "{}={}".format(k, repr(v)) for k, v in self.__dict__.items() if k not in skip
        )
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))



"""Action menu request message handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..base_service import BaseMenuService
from ..messages.menu_request import MenuRequest


class MenuRequestHandler(BaseHandler):
    """Message handler class for action menu requests."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for action menu requests.

        Args:
            context: request context
            responder: responder callback
        """
        self._logger.debug("MenuRequestHandler called with context %s", context)
        assert isinstance(context.message, MenuRequest)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        self._logger.info("Received action menu request")

        service: BaseMenuService = context.inject_or(BaseMenuService)
        if service:
            menu = await service.get_active_menu(
                context.profile,
                context.connection_record,
                context.message._thread_id,
            )
            if menu:
                await responder.send_reply(menu)
        else:
            # send problem report?
            pass
